import abjad

import nauert


def test_QGrid___call___01():
    q_grid = nauert.QGrid()
    a = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(0), ["A"]), abjad.Offset(0))
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 20), ["B"]), abjad.Offset(1, 20)
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(9, 20), ["C"]), abjad.Offset(9, 20)
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["D"]), abjad.Offset(1, 2)
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(11, 20), ["E"]), abjad.Offset(11, 20)
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(19, 20), ["F"]), abjad.Offset(19, 20)
    )
    g = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(1), ["G"]), abjad.Offset(1))
    q_grid.fit_q_events([a, b, c, d, e, f, g])
    result = q_grid(abjad.Duration(1, 4))
    assert len(result) == 1
    assert repr(result[0]) == """Note("c'4")"""

    annotation = abjad.get.indicator(result[0], dict)
    q_events = annotation["q_events"]
    assert isinstance(q_events, tuple) and len(q_events) == 4
    assert q_events[0].attachments == ("A",)
    assert q_events[1].attachments == ("B",)
    assert q_events[2].attachments == ("C",)
    assert q_events[3].attachments == ("D",)


def test_QGrid___call___02():
    q_grid = nauert.QGrid()
    q_grid.subdivide_leaves([(0, (1, 1, 1))])
    q_grid.subdivide_leaves([(1, (1, 1))])
    q_grid.subdivide_leaves([(-2, (1, 1, 1))])
    a = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(0), ["A"]), abjad.Offset(0))
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 20), ["B"]), abjad.Offset(1, 20)
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(9, 20), ["C"]), abjad.Offset(9, 20)
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["D"]), abjad.Offset(1, 2)
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(11, 20), ["E"]), abjad.Offset(11, 20)
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(19, 20), ["F"]), abjad.Offset(19, 20)
    )
    g = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(1), ["G"]), abjad.Offset(1))
    q_grid.fit_q_events([a, b, c, d, e, f, g])
    result = q_grid(abjad.Duration(1, 4))
    assert isinstance(result, list) and len(result) == 1
    assert abjad.lilypond(result[0]) == abjad.string.normalize(
        r"""
        \tuplet 3/2
        {
            c'8
            c'16
            c'16
            \tuplet 3/2
            {
                c'16
                c'16
                c'16
            }
        }
        """
    ), print(format(result[0]))

    leaves = abjad.select.leaves(result[0])
    leaf = leaves[0]
    annotation = abjad.get.indicator(leaf, dict)
    q_events = annotation["q_events"]
    assert isinstance(q_events, tuple) and len(q_events) == 2
    assert q_events[0].attachments == ("A",)
    assert q_events[1].attachments == ("B",)

    leaf = leaves[1]
    assert not abjad.get.indicator(leaf, dict)

    leaf = leaves[2]
    annotation = abjad.get.indicator(leaf, dict)
    q_events = annotation["q_events"]
    assert isinstance(q_events, tuple) and len(q_events) == 3
    assert q_events[0].attachments == ("C",)
    assert q_events[1].attachments == ("D",)
    assert q_events[2].attachments == ("E",)
    for leaf in leaves[3:6]:
        assert not abjad.get.indicator(leaf, dict)


def test_QGrid___call___03():
    """
    Non-binary works too.
    """
    q_grid = nauert.QGrid()
    q_grid.subdivide_leaves([(0, (1, 1))])
    a = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(0), ["A"]), abjad.Offset(0))
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 20), ["B"]), abjad.Offset(1, 20)
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(9, 20), ["C"]), abjad.Offset(9, 20)
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["D"]), abjad.Offset(1, 2)
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(11, 20), ["E"]), abjad.Offset(11, 20)
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(19, 20), ["F"]), abjad.Offset(19, 20)
    )
    g = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(1), ["G"]), abjad.Offset(1))
    q_grid.fit_q_events([a, b, c, d, e, f, g])
    result = q_grid(abjad.Duration(1, 3))
    assert isinstance(result, list) and len(result) == 1
    assert abjad.lilypond(result[0]) == abjad.string.normalize(
        r"""
        \tweak edge-height #'(0.7 . 0)
        \tuplet 3/2
        {
            c'4
            c'4
        }
        """
    ), print(format(result[0]))



import abjad

import nauert


def test_BeatwiseQSchema___init___01():
    item_a = nauert.BeatwiseQSchemaItem(
        search_tree=nauert.UnweightedSearchTree({2: None})
    )
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 76)
    item_b = nauert.BeatwiseQSchemaItem(tempo=metronome_mark)
    item_c = nauert.BeatwiseQSchemaItem(beatspan=abjad.Duration(1, 8))

    schema = nauert.BeatwiseQSchema(
        {2: item_a, 4: item_b, 7: item_c},
        beatspan=abjad.Duration(1, 32),
        search_tree=nauert.UnweightedSearchTree({3: None}),
        tempo=abjad.MetronomeMark(abjad.Duration(1, 16), 32),
    )
    assert len(schema.items) == 3
    assert schema.beatspan == abjad.Duration(1, 32)
    assert schema.search_tree == nauert.UnweightedSearchTree({3: None})
    assert schema.tempo == abjad.MetronomeMark(abjad.Duration(1, 16), 32)


def test_BeatwiseQSchema___init___02():
    schema = nauert.BeatwiseQSchema()
    assert len(schema.items) == 0
    assert schema.beatspan == abjad.Duration(1, 4)
    assert schema.search_tree == nauert.UnweightedSearchTree()
    assert schema.tempo == abjad.MetronomeMark(abjad.Duration(1, 4), 60)



import bisect
import copy
import typing

import abjad
import uqbar.containers
import uqbar.graphs

from . import qeventproxy as _qeventproxy


class QGridLeaf(abjad.rhythmtrees.RhythmTreeNode, uqbar.containers.UniqueTreeNode):
    """
    Q-grid leaf.

    ..  container:: example

        >>> nauert.QGridLeaf()
        QGridLeaf((1, 1), q_event_proxies=[], is_divisible=True)

    Used internally by ``QGrid``.
    """

    ### INITIALIZER ###

    def __init__(
        self,
        preprolated_duration: abjad.Duration = abjad.Duration(1),
        q_event_proxies: typing.Sequence[_qeventproxy.QEventProxy] = (),
        is_divisible: bool = True,
    ) -> None:
        assert isinstance(preprolated_duration, abjad.Duration), repr(
            preprolated_duration
        )
        assert q_event_proxies is not None, repr(q_event_proxies)
        assert all(isinstance(x, _qeventproxy.QEventProxy) for x in q_event_proxies)
        assert isinstance(is_divisible, bool), repr(is_divisible)
        uqbar.containers.UniqueTreeNode.__init__(self)
        abjad.rhythmtrees.RhythmTreeNode.__init__(self, preprolated_duration.pair)
        self._q_event_proxies = list(q_event_proxies)
        self._is_divisible = is_divisible

    ### SPECIAL METHODS ###

    def __call__(
        self, pulse_duration: abjad.Duration
    ) -> list[abjad.Note | abjad.Tuplet]:
        """
        Calls q-grid leaf.
        """
        assert isinstance(pulse_duration, abjad.Duration), repr(pulse_duration)
        total_duration = pulse_duration * abjad.Duration(self.pair)
        return abjad.makers.make_notes(0, total_duration)

    def __graph__(self, **keywords: None) -> uqbar.graphs.Graph:
        """
        Graphviz graph of q-grid leaf.
        """
        graph = uqbar.graphs.Graph(name="G")
        label = str(self.pair)
        node = uqbar.graphs.Node(attributes={"label": label, "shape": "box"})
        graph.append(node)
        return graph

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}({self.pair!r},"
        string += f" q_event_proxies={self.q_event_proxies!r},"
        string += f" is_divisible={self.is_divisible!r})"
        return string

    ### PRIVATE PROPERTIES ###

    @property
    def _pretty_rtm_format_pieces(self) -> list[str]:
        numerator, denominator = self.pair
        assert denominator == 1, repr(denominator)
        return [str(numerator)]

    ### PUBLIC PROPERTIES ###

    @property
    def is_divisible(self) -> bool:
        """
        Flag for whether the node may be further divided
        under some search tree.
        """
        return self._is_divisible

    @is_divisible.setter
    def is_divisible(self, argument):
        self._is_divisible = bool(argument)

    @property
    def preceding_q_event_proxies(self) -> list[_qeventproxy.QEventProxy]:
        """
        Gets preceding q-event proxies of q-grid leaf.
        """
        return [x for x in self._q_event_proxies if x.offset < self.start_offset]

    @property
    def q_event_proxies(self) -> list[_qeventproxy.QEventProxy]:
        """
        Gets q-event proxies of q-grid leaf.
        """
        return self._q_event_proxies

    @property
    def rtm_format(self) -> str:
        """
        Gets RTM format of q-grid leaf.
        """
        numerator, denominator = self.pair
        assert denominator == 1, repr(denominator)
        return str(numerator)

    @property
    def succeeding_q_event_proxies(self) -> list[_qeventproxy.QEventProxy]:
        """
        Gets succeeding q-event proxies of q-grid leaf.
        """
        return [x for x in self._q_event_proxies if self.start_offset <= x.offset]


class QGridContainer(abjad.rhythmtrees.RhythmTreeContainer):
    """
    Q-grid container.

    ..  container:: example

        >>> nauert.QGridContainer((1, 1))
        QGridContainer((1, 1))

    Used internally by ``QGrid``.
    """

    ### PRIVATE PROPERTIES ###

    @property
    def _leaf_class(self) -> type:
        return QGridLeaf

    @property
    def _node_class(self) -> tuple[type, type]:
        return (type(self), QGridLeaf)

    ### PUBLIC PROPERTIES ###

    @property
    def leaves(self) -> tuple[QGridLeaf, ...]:
        """
        Gets leaves.
        """
        return tuple(_ for _ in self.depth_first() if isinstance(_, QGridLeaf))


class QGrid:
    """
    Q-grid.

    Rhythm-tree-based model for how millisecond attack points collapse onto the
    offsets generated by a nested rhythmic structure.

    ..  container:: example

        ``QGrids`` model not only the internal nodes of the nesting structure,
        but also the downbeat to the "next" ``QGrid``, allowing events which
        occur very late within one structure to collapse virtually onto the
        beginning of the next structure.

        ``QEventProxies`` can be "loaded in" to the node contained by the
        ``QGrid`` closest to their virtual offset:

        >>> q_grid = nauert.QGrid()
        >>> q_event_a = nauert.PitchedQEvent(abjad.Offset(250), [0])
        >>> q_event_b = nauert.PitchedQEvent(abjad.Offset(750), [1])
        >>> proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.25))
        >>> proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.75))
        >>> q_grid.fit_q_events([proxy_a, proxy_b])

        >>> for q_event_proxy in q_grid.root_node.q_event_proxies:
        ...     q_event_proxy
        ...
        QEventProxy(...)

        >>> for q_event_proxy in q_grid.next_downbeat.q_event_proxies:
        ...     q_event_proxy
        ...
        QEventProxy(...)

    Used internally by the ``quantize`` function.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_next_downbeat", "_root_node")

    ### INITIALIZATION ###

    def __init__(
        self,
        root_node: QGridLeaf | QGridContainer | None = None,
        next_downbeat: QGridLeaf | None = None,
    ) -> None:
        if root_node is None:
            root_node = QGridLeaf(abjad.Duration(1, 1))
        assert isinstance(
            root_node,
            (QGridLeaf, QGridContainer),
        )
        if next_downbeat is None:
            next_downbeat = QGridLeaf(abjad.Duration(1, 1))
        assert isinstance(next_downbeat, QGridLeaf)
        self._root_node = root_node
        self._next_downbeat = next_downbeat
        self._next_downbeat._offset = abjad.Offset(1)
        self._next_downbeat._offsets_are_current = True

    ### SPECIAL METHODS ###

    def __call__(self, beatspan: abjad.Duration) -> list[abjad.Leaf | abjad.Tuplet]:
        """
        Calls q-grid.
        """
        assert isinstance(beatspan, abjad.Duration), repr(beatspan)
        result = []
        components = self.root_node(beatspan)
        voice = abjad.Voice(components)
        for tuplet in abjad.select.components(voice, abjad.Tuplet):
            assert isinstance(tuplet, abjad.Tuplet)
            if tuplet.trivial():
                abjad.mutate.extract(tuplet)
        components_ = abjad.mutate.eject_contents(voice)
        for component in components_:
            assert isinstance(component, abjad.Leaf | abjad.Tuplet)
            result.append(component)
        result_logical_ties = list(abjad.iterate.logical_ties(result))
        for logical_tie, q_grid_leaf in zip(result_logical_ties, self.leaves[:-1]):
            if q_grid_leaf.q_event_proxies:
                q_events = [
                    q_event_proxy.q_event
                    for q_event_proxy in q_grid_leaf.q_event_proxies
                ]
                q_events.sort(
                    key=lambda x: 0 if x is None or x.index is None else x.index
                )
                annotation = {"q_events": tuple(q_events)}
                leaf = abjad.get.leaf(logical_tie, 0)
                abjad.attach(annotation, leaf)
        return result

    def __copy__(self, *arguments: None) -> "QGrid":
        """
        Copies q-grid.
        """
        root_node, next_downbeat = self._root_node, self._next_downbeat
        return type(self)(copy.deepcopy(root_node), copy.deepcopy(next_downbeat))

    def __eq__(self, argument) -> bool:
        """
        Is true if `argument` is a q-grid with root node and next downbeat
        equal to those of this q-grid. Otherwise false.
        """
        if type(self) is type(argument):
            if self.root_node == argument.root_node:
                if self.next_downbeat == argument.next_downbeat:
                    return True
        return False

    def __hash__(self) -> int:
        """
        Hashes q-grid.

        Required to be explicitly redefined on Python 3 if __eq__ changes.
        """
        return super(QGrid, self).__hash__()

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(root_node={self.root_node!r},"
        string += f" next_downbeat={self.next_downbeat!r})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def distance(self) -> typing.Optional[abjad.Duration]:
        r"""
        The computed total distance (divided by the number of ``QEventProxy``
        objects) of the offset of each ``QEventProxy`` contained by the
        ``QGrid`` to the offset of the ``QGridLeaf`` to which the
        ``QEventProxy`` is attached.

        ..  container:: example

            >>> q_grid = nauert.QGrid()
            >>> q_event_a = nauert.PitchedQEvent(abjad.Offset(250), [0], ["A"])
            >>> q_event_b = nauert.PitchedQEvent(abjad.Offset(750), [1], ["B"])
            >>> proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.25))
            >>> proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.75))
            >>> q_grid.fit_q_events([proxy_a, proxy_b])
            >>> print(q_grid.rtm_format)
            1

            >>> pairs = zip(q_grid.leaves, q_grid.offsets, strict=True)
            >>> for index, (leaf, offset) in enumerate(pairs):
            ...     for q_event_proxy in leaf.q_event_proxies:
            ...         q_event = q_event_proxy.q_event
            ...         print(
            ...             "leaf's index: {}, leaf's offset: {}, q_event: {}".format(
            ...                 index, offset, q_event.attachments
            ...             )
            ...         )
            ...
            leaf's index: 0, leaf's offset: 0, q_event: ('A',)
            leaf's index: 1, leaf's offset: 1, q_event: ('B',)

            >>> q_grid.distance
            Duration(1, 4)

            >>> q_events = q_grid.subdivide_leaves([(0, (1, 1))])
            >>> q_grid.fit_q_events(q_events)
            >>> q_events = q_grid.subdivide_leaves([(0, (1, 1))])
            >>> q_grid.fit_q_events(q_events)
            >>> print(q_grid.rtm_format)
            (1 ((1 (1 1)) 1))

            >>> pairs = zip(q_grid.leaves, q_grid.offsets, strict=True)
            >>> for index, (leaf, offset) in enumerate(pairs):
            ...     for q_event_proxy in leaf.q_event_proxies:
            ...         q_event = q_event_proxy.q_event
            ...         print(
            ...             "leaf's index: {}, leaf's offset: {}, q_event: {}".format(
            ...                 index, offset, q_event.attachments
            ...             )
            ...         )
            ...
            leaf's index: 1, leaf's offset: 1/4, q_event: ('A',)
            leaf's index: 2, leaf's offset: 1/2, q_event: ('B',)

            >>> q_grid.distance
            Duration(1, 8)

        """
        count = 0
        absolute_distance = abjad.Duration(0)
        for leaf, offset in zip(self.leaves, self.offsets):
            for q_event_proxy in leaf.q_event_proxies:
                absolute_distance += abs(q_event_proxy.offset - offset)
                count += 1
        if count:
            return absolute_distance / count
        return None

    @property
    def leaves(self) -> tuple[QGridLeaf, ...]:
        """
        Gets all of the leaf nodes in the QGrid, including the next downbeat's
        node.
        """
        if isinstance(self._root_node, QGridLeaf):
            return (self._root_node, self._next_downbeat)
        return self._root_node.leaves + (self._next_downbeat,)

    @property
    def next_downbeat(self) -> QGridLeaf:
        """
        Gets the node representing the "next" downbeat after the contents of
        the QGrid's tree.
        """
        return self._next_downbeat

    @property
    def offsets(self) -> tuple[abjad.Offset, ...]:
        """
        Gets the offsets between 0 and 1 of all of the leaf nodes in the QGrid.
        """
        return tuple([x.start_offset for x in self.leaves[:-1]] + [abjad.Offset(1)])

    @property
    def pretty_rtm_format(self) -> str:
        """
        Gets the pretty RTM-format of the root node of the ``QGrid``.
        """
        return self._root_node.pretty_rtm_format

    @property
    def root_node(self) -> QGridLeaf | QGridContainer:
        """
        Gets the root node of the ``QGrid``.
        """
        return self._root_node

    @property
    def rtm_format(self) -> str:
        """
        Gets the RTM format of the root node of the ``QGrid``.
        """
        return self._root_node.rtm_format

    ### PUBLIC METHODS ###

    def fit_q_events(
        self, q_event_proxies: typing.Sequence[_qeventproxy.QEventProxy]
    ) -> None:
        """
        Fits each ``QEventProxy`` in ``q_event_proxies`` onto the contained
        ``QGridLeaf`` whose offset is nearest.
        """
        assert all(isinstance(x, _qeventproxy.QEventProxy) for x in q_event_proxies)
        leaves, offsets = self.leaves, self.offsets
        for q_event_proxy in q_event_proxies:
            idx = bisect.bisect_left(offsets, q_event_proxy.offset)
            if q_event_proxy.offset == offsets[idx]:
                leaves[idx].q_event_proxies.append(q_event_proxy)
            else:
                left, right = offsets[idx - 1], offsets[idx]
                left_diff = abs(left - q_event_proxy.offset)
                right_diff = abs(right - q_event_proxy.offset)
                if right_diff < left_diff:
                    leaves[idx].q_event_proxies.append(q_event_proxy)
                else:
                    leaves[idx - 1].q_event_proxies.append(q_event_proxy)

    def regroup_leaves_with_unencessary_divisions(self) -> None:
        """
        Regroups leaves that belong to the same parent in which only the first
        leaf contains q_event_prox[y|ies].
        """
        index = 0
        while True:
            leaf = self.leaves[index]
            parent = leaf.parent
            if isinstance(parent, QGridContainer):
                leaves = parent.leaves
                if len(leaves) > 1 and all(
                    [_leaf.q_event_proxies == [] for _leaf in leaves[1:]]
                ):
                    new_leaf = QGridLeaf(
                        preprolated_duration=abjad.Duration(parent.pair),
                        q_event_proxies=leaves[0].q_event_proxies,
                    )
                    index = parent.parent.index(parent)
                    parent.parent[index] = [new_leaf]
            index += 1
            if index == len(self.leaves):
                break

    def sort_q_events_by_index(self) -> None:
        """
        Sorts ``QEventProxies`` attached to each ``QGridLeaf`` in a ``QGrid``
        by their index.
        """
        for leaf in self.leaves:
            leaf.q_event_proxies.sort(key=lambda x: 0 if x.index is None else x.index)

    def subdivide_leaf(
        self,
        leaf: QGridLeaf,
        subdivisions: typing.Sequence[abjad.Duration | int],
    ) -> list[_qeventproxy.QEventProxy]:
        """
        Replaces the QGridLeaf ``leaf`` contained in a QGrid by a
        QGridContainer containing QGridLeaves with durations equal to the ratio
        described in ``subdivisions.``

        Returns list of QEventProxies attached to ``leaf``.
        """
        children = []
        for subdivision in subdivisions:
            if isinstance(subdivision, int):
                subdivision = abjad.Duration(subdivision)
                child = QGridLeaf(preprolated_duration=subdivision)
                children.append(child)
            else:
                assert isinstance(subdivision, abjad.Duration), repr(subdivision)
        container = QGridContainer(
            leaf.pair,
            children=[
                QGridLeaf(preprolated_duration=abjad.Duration(subdivision))
                for subdivision in subdivisions
            ],
        )
        if leaf.parent is not None:
            index = leaf.parent.index(leaf)
            leaf.parent[index] = [container]
        # otherwise, our root node if just a QGridLeaf
        else:
            self._root_node = container
        return leaf.q_event_proxies

    def subdivide_leaves(
        self, pairs: typing.Sequence[tuple[int, tuple[int, int]]]
    ) -> list[_qeventproxy.QEventProxy]:
        r"""
        Given a sequence of leaf-index:subdivision-ratio pairs ``pairs``,
        subdivide the ``QGridLeaves`` described by the indices into
        ``QGridContainers`` containing ``QGridLeaves`` with durations equal to
        their respective subdivision-ratios.

        Returns the ``QEventProxies`` attached to thus subdivided
        ``QGridLeaf``.
        """
        pairs = sorted(dict(pairs).items())
        leaf_indices = [pair[0] for pair in pairs]
        subdivisions = [pair[1] for pair in pairs]
        all_leaves = self.leaves
        leaves_to_subdivide = [all_leaves[idx] for idx in leaf_indices]
        q_event_proxies = []
        for i, leaf in enumerate(leaves_to_subdivide):
            next_leaf = all_leaves[all_leaves.index(leaf) + 1]
            if next_leaf is self.next_downbeat:
                next_leaf_offset = abjad.Offset(1)
            else:
                next_leaf_offset = next_leaf.start_offset

            q_event_proxies.extend(self.subdivide_leaf(leaf, subdivisions[i]))
            for q_event_proxy in tuple(next_leaf.q_event_proxies):
                if q_event_proxy.offset < next_leaf_offset:
                    idx = next_leaf.q_event_proxies.index(q_event_proxy)
                    q_event_proxies.append(next_leaf.q_event_proxies.pop(idx))
        return q_event_proxies



import abjad

import nauert


def test_QGrid___eq___01():
    a = nauert.QGrid()
    b = nauert.QGrid()
    assert format(a) == format(b)
    assert a != b


def test_QGrid___eq___02():
    a = nauert.QGrid(
        root_node=nauert.QGridContainer(
            (1, 1),
            children=[
                nauert.QGridLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    q_event_proxies=[
                        nauert.QEventProxy(
                            nauert.SilentQEvent(abjad.Offset(100)),
                            abjad.Offset(0.5),
                        )
                    ],
                )
            ],
        ),
        next_downbeat=nauert.QGridLeaf(
            preprolated_duration=abjad.Duration(1, 1),
            q_event_proxies=[
                nauert.QEventProxy(
                    nauert.TerminalQEvent(abjad.Offset(200)), abjad.Offset(0.9)
                )
            ],
        ),
    )
    b = nauert.QGrid(
        root_node=nauert.QGridContainer(
            (1, 1),
            children=[
                nauert.QGridLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    q_event_proxies=[
                        nauert.QEventProxy(
                            nauert.SilentQEvent(abjad.Offset(100)),
                            abjad.Offset(0.5),
                        )
                    ],
                )
            ],
        ),
        next_downbeat=nauert.QGridLeaf(
            preprolated_duration=abjad.Duration(1, 1),
            q_event_proxies=[
                nauert.QEventProxy(
                    nauert.TerminalQEvent(abjad.Offset(200)), abjad.Offset(0.9)
                )
            ],
        ),
    )
    assert format(a) == format(b)
    assert a != b


def test_QGrid___eq___03():
    a = nauert.QGrid()
    b = nauert.QGrid(
        root_node=nauert.QGridContainer(
            (1, 1),
            children=[
                nauert.QGridLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    q_event_proxies=[
                        nauert.QEventProxy(
                            nauert.SilentQEvent(abjad.Offset(100)),
                            abjad.Offset(0.5),
                        )
                    ],
                )
            ],
        )
    )
    c = nauert.QGrid(
        next_downbeat=nauert.QGridLeaf(
            preprolated_duration=abjad.Duration(1, 1),
            q_event_proxies=[
                nauert.QEventProxy(
                    nauert.TerminalQEvent(abjad.Offset(200)), abjad.Offset(0.9)
                )
            ],
        )
    )
    d = (
        nauert.QGrid(
            root_node=nauert.QGridContainer(
                (1, 1),
                children=[
                    nauert.QGridLeaf(
                        preprolated_duration=abjad.Duration(1, 1),
                        q_event_proxies=[
                            nauert.QEventProxy(
                                nauert.SilentQEvent(abjad.Offset(100)),
                                abjad.Offset(0.5),
                            )
                        ],
                    )
                ],
            ),
            next_downbeat=nauert.QGridLeaf(
                preprolated_duration=abjad.Duration(1, 1),
                q_event_proxies=[
                    nauert.QEventProxy(
                        nauert.TerminalQEvent(abjad.Offset(200)),
                        abjad.Offset(0.9),
                    )
                ],
            ),
        ),
    )
    assert a != b
    assert a != c
    assert a != d
    assert b != c
    assert b != d
    assert c != d



import abc
import typing

import abjad

from . import qevents as _qevents


class GraceHandler(abc.ABC):
    """
    Abstract grace-handler.

    Determines what pitch, if any, will be selected from a list of ``QEvents``
    to be applied to an attack-point generated by a ``QGrid``, and whether
    there should be a ``BeforeGraceContainer`` attached to that attack-point.

    When called on a sequence of ``QEvents``, ``GraceHandler`` subclasses
    should return a pair, where the first item of the pair is a sequence of
    pitch tokens or ``None``, and where the second item of the pair is a
    ``BeforeGraceContainer`` instance or None.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self):
        pass

    ### SPECIAL METHODS ###

    @abc.abstractmethod
    def __call__(self, q_events) -> tuple[
        tuple[abjad.NamedPitch, ...],
        typing.Optional[tuple],
        typing.Optional[abjad.BeforeGraceContainer],
    ]:
        """
        Calls grace handler.
        """
        raise NotImplementedError


class CollapsingGraceHandler(GraceHandler):
    r"""
    Collapsing grace-handler.

    Collapses pitch information into a single chord rather than creating a
    grace container.

    ..  container:: example

        >>> durations = [1000, 1, 1, 997]
        >>> pitches = [0, 7, 4, 0]
        >>> method = nauert.QEventSequence.from_millisecond_pitch_pairs
        >>> pairs = tuple(zip(durations, pitches, strict=True))
        >>> q_event_sequence = method(pairs)
        >>> grace_handler = nauert.CollapsingGraceHandler()
        >>> voice = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    <c' e' g'>4
                    r4
                    r4
                }
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __call__(
        self, q_events: typing.Sequence[_qevents.QEvent]
    ) -> tuple[tuple[abjad.NamedPitch, ...], tuple, None]:
        """
        Calls collapsing grace handler.
        """
        pitches: list[abjad.NamedPitch] = []
        attachments: list[typing.Any] = []
        for q_event in q_events:
            if isinstance(q_event, _qevents.PitchedQEvent):
                pitches.extend(q_event.pitches)
                attachments.extend(q_event.attachments)
        return tuple(pitches), tuple(attachments), None


class ConcatenatingGraceHandler(GraceHandler):
    r"""
    Concatenating grace-handler.

    Concatenates all but the final ``QEvent`` attached to a ``QGrid`` offset
    into a ``BeforeGraceContainer``, using a fixed leaf duration ``duration``.

    ..  container:: example

        When called, it returns pitch information of final ``QEvent``, and the
        generated ``BeforeGraceContainer``, if any:

        >>> durations = [1000, 1, 999]
        >>> pitches = [0, 2, 0]
        >>> pairs = tuple(zip(durations, pitches, strict=True))
        >>> q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(pairs)
        >>> grace_handler = nauert.ConcatenatingGraceHandler()
        >>> voice = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    \grace {
                        d'16
                    }
                    c'4
                    r4
                    r4
                }
            }

    ..  container:: example

        When ``discard_grace_rest`` is set to ``True`` (the default), all the
        grace rests are discarded:

        >>> durations = [1000, 1, 999]
        >>> pitches = [0, None, 0]
        >>> pairs = tuple(zip(durations, pitches, strict=True))
        >>> q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(pairs)
        >>> grace_handler = nauert.ConcatenatingGraceHandler()
        >>> voice = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    c'4
                    r4
                    r4
                }
            }

    ..  container:: example

        When ``discard_grace_rest`` is set to ``False``, grace rests are not
        discarded:

        >>> grace_handler = nauert.ConcatenatingGraceHandler(discard_grace_rest=False)
        >>> voice = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    \grace {
                        r16
                    }
                    c'4
                    r4
                    r4
                }
            }

    ..  container:: example

        When ``replace_rest_with_final_grace_note`` is set to ``False``, grace
        notes are allowed to be attached to a rest:

        >>> durations = [1000, 1, 999, 1000]
        >>> pitches = [0, 0, None, 0]
        >>> pairs = tuple(zip(durations, pitches, strict=True))
        >>> q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(pairs)
        >>> grace_handler = nauert.ConcatenatingGraceHandler(
        ...     replace_rest_with_final_grace_note=False
        ... )
        >>> voice = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    \grace {
                        c'16
                    }
                    r4
                    c'4
                    r4
                }
            }

    ..  container:: example

        When ``replace_rest_with_final_grace_note`` is set to ``True`` (the
        default behavior), any rest with grace notes attached to it is replaced
        by the last pitched grace note in the grace container:

        >>> grace_handler = nauert.ConcatenatingGraceHandler()
        >>> voice = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    c'4
                    c'4
                    r4
                }
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_discard_grace_rest",
        "_grace_duration",
        "_replace_rest_with_final_grace_note",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        discard_grace_rest: bool = True,
        grace_duration: abjad.Duration = abjad.Duration(1, 16),
        replace_rest_with_final_grace_note: bool = True,
    ) -> None:
        assert isinstance(discard_grace_rest, bool), repr(discard_grace_rest)
        assert isinstance(grace_duration, abjad.Duration), repr(grace_duration)
        assert isinstance(replace_rest_with_final_grace_note, bool)
        self._discard_grace_rest = discard_grace_rest
        assert grace_duration.is_dyadic_rational
        self._grace_duration = grace_duration
        self._replace_rest_with_final_grace_note = replace_rest_with_final_grace_note

    ### SPECIAL METHODS ###

    def __call__(self, q_events: typing.Sequence[_qevents.QEvent]) -> tuple[
        tuple[abjad.NamedPitch, ...],
        typing.Optional[tuple],
        typing.Optional[abjad.BeforeGraceContainer],
    ]:
        """
        Calls concatenating grace handler.
        """
        grace_events, final_event = q_events[:-1], q_events[-1]
        attachments: tuple | None
        if grace_events and self._replace_rest_with_final_grace_note:
            index = self._find_last_pitched_q_event(q_events)
            grace_events, final_event = q_events[:index], q_events[index]
        if isinstance(final_event, _qevents.PitchedQEvent):
            # TODO: we are only supporting preserving attachments for PitchedQEvent
            pitches = final_event.pitches
            attachments = final_event.attachments
        else:
            pitches = ()
            attachments = None
        grace_events_list = list(grace_events)
        if self._discard_grace_rest:
            for q_event in grace_events_list:
                if isinstance(q_event, _qevents.SilentQEvent):
                    grace_events_list.remove(q_event)
        grace_events = tuple(grace_events_list)
        grace_container: abjad.BeforeGraceContainer | None
        if grace_events:
            grace_container = abjad.BeforeGraceContainer()
            for q_event in grace_events:
                leaf: abjad.Leaf
                if isinstance(q_event, _qevents.PitchedQEvent):
                    if len(q_event.pitches) == 1:
                        leaf = abjad.Note(q_event.pitches[0], self.grace_duration)
                    else:
                        leaf = abjad.Chord(q_event.pitches, self.grace_duration)
                else:
                    leaf = abjad.Rest(self.grace_duration)
                q_event_attachments = (
                    None if not hasattr(q_event, "attachments") else q_event.attachments
                )
                if q_event_attachments is not None:
                    abjad.annotate(leaf, "q_event_attachments", q_event_attachments)
                grace_container.append(leaf)
        else:
            grace_container = None

        return tuple(pitches), attachments, grace_container

    ### PRIVATE METHODS ###

    @staticmethod
    def _find_last_pitched_q_event(q_events: typing.Sequence[_qevents.QEvent]) -> int:
        for index, q_event in enumerate(reversed(q_events)):
            if isinstance(q_event, _qevents.PitchedQEvent):
                return len(q_events) - index - 1
        raise ValueError("There should be at least one PitchedQEvent in q_events")

    ### PUBLIC METHODS ###

    def handle_orphaned_q_event_proxies(self, last_leaf, q_event_proxies):
        r"""
        Embeds orphaned ``QEvents`` into an ``AfterGraceContainer`` and
        attaches it to ``last_leaf``.

        ..  container:: example

            >>> durations = [1000, 1000, 1000, 400, 50, 50]
            >>> pitches = range(len(durations))
            >>> pairs = tuple(zip(durations, pitches, strict=True))
            >>> method = nauert.QEventSequence.from_millisecond_pitch_pairs
            >>> q_event_sequence = method(pairs)
            >>> search_tree = nauert.UnweightedSearchTree()
            >>> attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
            >>> q_schema = nauert.MeasurewiseQSchema(
            ...     search_tree=search_tree,
            ...     time_signature=(7, 8),
            ...     use_full_measure=True,
            ... )
            >>> result = nauert.quantize(
            ...     q_event_sequence,
            ...     q_schema=q_schema,
            ...     attach_tempos=True,
            ...     attack_point_optimizer=attack_point_optimizer,
            ... )
            >>> staff = abjad.Staff([result])
            >>> score = abjad.Score([staff], name="Score")
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    \new Voice
                    {
                        {
                            \tempo 4=60
                            \time 7/8
                            c'4
                            cs'8
                            ~
                            cs'8
                            d'8
                            ~
                            d'8
                            \afterGrace
                            ef'8
                            {
                                e'16
                                f'16
                            }
                        }
                    }
                }

        """
        grace_container = abjad.AfterGraceContainer() if q_event_proxies else None
        for proxy in q_event_proxies:
            q_event = proxy.q_event
            if isinstance(q_event, _qevents.PitchedQEvent):
                if len(q_event.pitches) == 1:
                    leaf = abjad.Note(q_event.pitches[0], self.grace_duration)
                else:
                    leaf = abjad.Chord(q_event.pitches, self.grace_duration)
                abjad.annotate(leaf, "q_event_attachments", q_event.attachments)
                assert grace_container is not None
                grace_container.append(leaf)
        if 0 < len(grace_container):
            abjad.attach(grace_container, last_leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def discard_grace_rest(self) -> bool:
        """
        Boolean of whether to discard grace rests or not.
        """
        return self._discard_grace_rest

    @property
    def grace_duration(self) -> abjad.Duration:
        """
        Grace duration of concantenating grace handler.
        """
        return self._grace_duration

    @property
    def replace_rest_with_final_grace_note(self) -> bool:
        """
        Boolean of whether to replace the rest with the final (pitched) grace
        note.
        """
        return self._replace_rest_with_final_grace_note


class DiscardingGraceHandler(GraceHandler):
    r"""
    Discarding grace-handler.

    Discards all but final q-event attached to an offset.

    Does not create grace containers.

    ..  container:: example

        >>> durations = [1000, 1, 1, 998]
        >>> pitches = [0, 1, 2, 3]
        >>> pairs = tuple(zip(durations, pitches, strict=True))
        >>> q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(pairs)
        >>> grace_handler = nauert.DiscardingGraceHandler()
        >>> voice = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
        >>> staff = abjad.Staff([voice])
        >>> score = abjad.Score([staff])
        >>> abjad.show(score) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(voice)
            >>> print(string)
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 4/4
                    c'4
                    ef'4
                    r4
                    r4
                }
            }

        >>> for pqe in grace_handler.discarded_q_events[0]:
        ...     pqe
        PitchedQEvent(offset=Offset((1000, 1)), pitches=(NamedPitch("cs'"),), ...)
        PitchedQEvent(offset=Offset((1001, 1)), pitches=(NamedPitch("d'"),), ...)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_discarded_q_events",)

    ### INITIALIZER ###

    def __init__(self) -> None:
        self._discarded_q_events: list[typing.Sequence[_qevents.QEvent]] = []

    ### SPECIAL METHODS ###

    def __call__(
        self, q_events: typing.Sequence[_qevents.QEvent]
    ) -> tuple[tuple[abjad.NamedPitch, ...], tuple, None]:
        """
        Calls discarding grace handler.
        """
        if q_events[:-1]:
            self._discarded_q_events.append(q_events[:-1])
        q_event = q_events[-1]
        if isinstance(q_event, _qevents.PitchedQEvent):
            return tuple(q_event.pitches), tuple(q_event.attachments), None
        return (), (), None

    @property
    def discarded_q_events(self) -> list[typing.Sequence[_qevents.QEvent]]:
        """
        Gets discarded QEvents.
        """
        return self._discarded_q_events



import abjad

import nauert


def test_QGrid_subdivide_leaves_01():
    q_grid = nauert.QGrid()
    a = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(0), ["A"]), abjad.Offset(0))
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 20), ["B"]), abjad.Offset(1, 20)
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(9, 20), ["C"]), abjad.Offset(9, 20)
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["D"]), abjad.Offset(1, 2)
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(11, 20), ["E"]), abjad.Offset(11, 20)
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(19, 20), ["F"]), abjad.Offset(19, 20)
    )
    g = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(1), ["G"]), abjad.Offset(1))
    q_grid.leaves[0].q_event_proxies.extend([a, b, c, d])
    q_grid.leaves[1].q_event_proxies.extend([e, f, g])
    assert q_grid.root_node.rtm_format == "1"

    q_events = q_grid.subdivide_leaves([(0, (1, 1))])
    assert q_events == [a, b, c, d, e, f]
    assert q_grid.root_node.rtm_format == "(1 (1 1))"

    q_grid.leaves[0].q_event_proxies.extend([a, b])
    q_grid.leaves[1].q_event_proxies.extend([c, d, e])
    q_grid.leaves[2].q_event_proxies.append(f)
    q_events = q_grid.subdivide_leaves([(0, (1, 1)), (1, (1, 1))])
    assert q_events == [a, b, c, d, e, f]
    assert q_grid.root_node.rtm_format == "(1 ((1 (1 1)) (1 (1 1))))"


def test_QGrid_subdivide_leaves_02():
    q_grid = nauert.QGrid()
    a = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(0), ["A"]), abjad.Offset(0))
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 20), ["B"]), abjad.Offset(1, 20)
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(9, 20), ["C"]), abjad.Offset(9, 20)
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["D"]), abjad.Offset(1, 2)
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(11, 20), ["E"]), abjad.Offset(11, 20)
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(19, 20), ["F"]), abjad.Offset(19, 20)
    )
    g = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(1), ["G"]), abjad.Offset(1))
    q_grid.leaves[0].q_event_proxies.extend([a, b, c, d])
    q_grid.leaves[1].q_event_proxies.extend([e, f, g])
    assert q_grid.root_node.rtm_format == "1"

    q_events = q_grid.subdivide_leaves([(0, (1, 1))])
    assert q_events == [a, b, c, d, e, f]
    assert q_grid.root_node.rtm_format == "(1 (1 1))"

    q_grid.leaves[0].q_event_proxies.extend([a, b])
    q_grid.leaves[1].q_event_proxies.extend([c, d, e])
    q_grid.leaves[2].q_event_proxies.append(f)
    q_events = q_grid.subdivide_leaves([(0, (3, 4, 5))])
    assert q_events == [a, b, c]
    assert q_grid.root_node.rtm_format == "(1 ((1 (3 4 5)) 1))"



import abjad

import nauert


def assert_q_event_attachments(result, all_attachments):
    for logical_tie, attachments in zip(
        abjad.iterate.logical_ties(result, pitched=True), all_attachments
    ):
        first_leaf = abjad.get.leaf(logical_tie, 0)
        q_event_attachments = abjad.get.annotation(first_leaf, "q_event_attachments")
        assert q_event_attachments == attachments, print(q_event_attachments)


def test_quantize_01():
    milliseconds = [1500, 1500]
    q_events = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    result = nauert.quantize(q_events)
    staff = abjad.Staff([result], lilypond_type="RhythmicStaff")
    score = abjad.Score([staff])
    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new RhythmicStaff
            {
                \new Voice
                {
                    {
                        \tempo 4=60
                        \time 4/4
                        c'4.
                        c'4.
                        r4
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_02():
    milliseconds = [750, 750]
    q_events = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    optimizer = nauert.MeasurewiseAttackPointOptimizer()
    result = nauert.quantize(q_events, attack_point_optimizer=optimizer)
    staff = abjad.Staff([result], lilypond_type="RhythmicStaff")
    score = abjad.Score([staff])
    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new RhythmicStaff
            {
                \new Voice
                {
                    {
                        \tempo 4=60
                        \time 4/4
                        c'8.
                        c'16
                        ~
                        c'8
                        r8
                        r2
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_03():
    milliseconds = [1500, -1000, 1000, 1000, -1000, 1000, -1000, 500]
    sequence = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    attack_point_optimizer = nauert.NullAttackPointOptimizer()

    result = nauert.quantize(sequence, attack_point_optimizer=attack_point_optimizer)

    assert isinstance(result, abjad.Voice)
    assert abjad.get.duration(result) == 2

    score = abjad.Score([abjad.Staff([result])])

    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new Staff
            {
                \new Voice
                {
                    {
                        \tempo 4=60
                        \time 4/4
                        c'4
                        ~
                        c'8
                        r8
                        r8
                        c'8
                        ~
                        c'8
                        c'8
                        ~
                    }
                    {
                        c'8
                        r8
                        r8
                        c'8
                        ~
                        c'8
                        r8
                        r8
                        c'8
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_04():
    milliseconds = [250, 1000, 1000, 1000, 750]
    sequence = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    attack_point_optimizer = nauert.NullAttackPointOptimizer()
    result = nauert.quantize(sequence, attack_point_optimizer=attack_point_optimizer)
    staff = abjad.Staff([result], lilypond_type="RhythmicStaff")
    score = abjad.Score([staff])

    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new RhythmicStaff
            {
                \new Voice
                {
                    {
                        \tempo 4=60
                        \time 4/4
                        c'16
                        c'16
                        ~
                        c'8
                        ~
                        c'16
                        c'16
                        ~
                        c'8
                        ~
                        c'16
                        c'16
                        ~
                        c'8
                        ~
                        c'16
                        c'16
                        ~
                        c'8
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_05():
    q_schema = nauert.BeatwiseQSchema(
        {"search_tree": nauert.UnweightedSearchTree({2: None})},
        {"search_tree": nauert.UnweightedSearchTree({3: None})},
        {"search_tree": nauert.UnweightedSearchTree({5: None})},
        {"search_tree": nauert.UnweightedSearchTree({7: None})},
    )
    milliseconds = [250, 250, 250, 250] * 4
    q_events = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    attack_point_optimizer = nauert.NullAttackPointOptimizer()

    result = nauert.quantize(
        q_events,
        q_schema=q_schema,
        attack_point_optimizer=attack_point_optimizer,
    )
    staff = abjad.Staff([result], lilypond_type="RhythmicStaff")
    score = abjad.Score([staff])

    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new RhythmicStaff
            {
                \new Voice
                {
                    \grace {
                        c'16
                    }
                    \tempo 4=60
                    c'8
                    \grace {
                        c'16
                    }
                    c'8
                    \tuplet 3/2
                    {
                        c'8
                        \grace {
                            c'16
                        }
                        c'8
                        c'8
                    }
                    \tuplet 5/4
                    {
                        c'16
                        c'16
                        c'16
                        ~
                        c'16
                        c'16
                    }
                    \tuplet 7/4
                    {
                        c'16
                        ~
                        c'16
                        c'16
                        c'16
                        ~
                        c'16
                        c'16
                        ~
                        c'16
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_06():
    milliseconds = [1000] * 8
    sequence = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    optimizer = nauert.MeasurewiseAttackPointOptimizer()
    result = nauert.quantize(sequence, attack_point_optimizer=optimizer)
    staff = abjad.Staff([result], lilypond_type="RhythmicStaff")
    score = abjad.Score([staff])

    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new RhythmicStaff
            {
                \new Voice
                {
                    {
                        \tempo 4=60
                        \time 4/4
                        c'4
                        c'4
                        c'4
                        c'4
                    }
                    {
                        c'4
                        c'4
                        c'4
                        c'4
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_07():
    milliseconds = [1000, 750, 1000, 1250] * 2
    sequence = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    optimizer = nauert.MeasurewiseAttackPointOptimizer()
    result = nauert.quantize(sequence, attack_point_optimizer=optimizer)
    staff = abjad.Staff([result], lilypond_type="RhythmicStaff")
    score = abjad.Score([staff])

    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new RhythmicStaff
            {
                \new Voice
                {
                    {
                        \tempo 4=60
                        \time 4/4
                        c'4
                        c'8.
                        c'16
                        ~
                        c'8.
                        c'16
                        ~
                        c'4
                    }
                    {
                        c'4
                        c'8.
                        c'16
                        ~
                        c'8.
                        c'16
                        ~
                        c'4
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_08():
    q_schema = nauert.BeatwiseQSchema(
        {"search_tree": nauert.UnweightedSearchTree({2: None})},
        {"search_tree": nauert.UnweightedSearchTree({3: None})},
        {"search_tree": nauert.UnweightedSearchTree({5: None})},
        {"search_tree": nauert.UnweightedSearchTree({7: None})},
    )
    milliseconds = [250, 250, 250, 250] * 4
    q_events = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    attack_point_optimizer = nauert.NaiveAttackPointOptimizer()

    result = nauert.quantize(
        q_events,
        q_schema=q_schema,
        attack_point_optimizer=attack_point_optimizer,
    )
    staff = abjad.Staff([result], lilypond_type="RhythmicStaff")
    score = abjad.Score([staff])

    assert abjad.lilypond(score) == abjad.string.normalize(
        r"""
        \new Score
        <<
            \new RhythmicStaff
            {
                \new Voice
                {
                    \grace {
                        c'16
                    }
                    \tempo 4=60
                    c'8
                    \grace {
                        c'16
                    }
                    c'8
                    \tuplet 3/2
                    {
                        c'8
                        \grace {
                            c'16
                        }
                        c'8
                        c'8
                    }
                    \tuplet 5/4
                    {
                        c'16
                        c'16
                        c'8
                        c'16
                    }
                    \tuplet 7/4
                    {
                        c'8
                        c'16
                        c'8
                        c'8
                    }
                }
            }
        >>
        """
    ), print(abjad.lilypond(score))


def test_Quantize_09():
    q_schema = nauert.BeatwiseQSchema(
        {"search_tree": nauert.UnweightedSearchTree({2: None})},
        {"search_tree": nauert.UnweightedSearchTree({3: None})},
        {"search_tree": nauert.UnweightedSearchTree({5: None})},
        {"search_tree": nauert.UnweightedSearchTree({7: None})},
    )
    milliseconds = [250, 250, 250, 250] * 4
    q_events = nauert.QEventSequence.from_millisecond_durations(milliseconds)
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()

    try:
        nauert.quantize(
            q_events,
            q_schema=q_schema,
            attack_point_optimizer=attack_point_optimizer,
        )
        assert False
    except TypeError as error:
        string = "BeatwiseQTarget is not supposed to be used together"
        string += " with MeasurewiseAttackPointOptimizer."
        assert str(error) == string


def test_Quantize_10():
    durations = [1000, 1000, 1000, 2000, 1000, 1000, 500, 500]
    pitches = range(8)
    all_attachments = [(x,) for x in pitches]
    pitches = [(x, x + 7) for x in pitches]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_attachment_tuples(
        tuple(zip(durations, pitches, all_attachments))
    )
    grace_handler = nauert.ConcatenatingGraceHandler(
        replace_rest_with_final_grace_note=True
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    string = abjad.lilypond(result)
    assert string == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                <c' g'>4
                <cs' af'>4
                <d' a'>4
                <ef' bf'>4
                ~
            }
            {
                <ef' bf'>4
                <e' b'>4
                <f' c''>4
                <fs' cs''>8
                <g' d''>8
            }
        }
        """
    ), print(string)
    assert_q_event_attachments(result, all_attachments)


def test_Quantize_11():
    durations = [250, 1250, 750, 1000, 250]
    pitches = range(5)
    all_attachments = [(x,) for x in pitches]
    pitches = [(x, x + 7) for x in pitches]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_attachment_tuples(
        tuple(zip(durations, pitches, all_attachments))
    )
    time_signature = abjad.TimeSignature((7, 8))
    search_tree = nauert.UnweightedSearchTree(
        definition={
            2: {2: None, 3: None},
            3: {2: None, 3: None},
            5: {2: None},
            7: {2: None},
            13: None,
        }
    )
    q_schema = nauert.MeasurewiseQSchema(
        time_signature=time_signature,
        search_tree=search_tree,
        use_full_measure=True,
    )
    grace_handler = nauert.ConcatenatingGraceHandler(
        replace_rest_with_final_grace_note=True
    )
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
    result = nauert.quantize(
        q_event_sequence,
        q_schema=q_schema,
        grace_handler=grace_handler,
        attack_point_optimizer=attack_point_optimizer,
    )
    staff = abjad.Staff([result])
    abjad.Score([staff], name="Score")
    string = abjad.lilypond(staff)
    assert string == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 7/8
                    <c' g'>16
                    <cs' af'>16
                    ~
                    <cs' af'>4
                    <d' a'>8.
                    <ef' bf'>16
                    ~
                    <ef' bf'>8.
                    <e' b'>16
                }
            }
        }
        """
    ), print(string)
    assert_q_event_attachments(result, all_attachments)


def test_Quantize_12():
    definition = {"divisors": (2, 3, 5, 7), "max_depth": 2, "max_divisions": 2}
    search_tree = nauert.WeightedSearchTree(definition=definition)
    durations = [457.14285, 814.1, 228.5714, 1440, 960]
    pitches = range(len(durations))
    all_attachments = [(x,) for x in pitches]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_attachment_tuples(
        tuple(zip(durations, pitches, all_attachments))
    )
    q_schema = nauert.MeasurewiseQSchema(
        search_tree=search_tree, time_signature=(7, 8), use_full_measure=True
    )
    result = nauert.quantize(q_event_sequence, q_schema=q_schema, attach_tempos=True)
    staff = abjad.Staff([result])
    abjad.Score([staff], name="Score")
    string = abjad.lilypond(staff)
    assert string == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \tweak text #tuplet-number::calc-fraction-text
                    \tuplet 7/6
                    {
                        \tempo 4=60
                        \time 7/8
                        c'8
                        cs'4
                        ~
                        cs'16
                    }
                    \tuplet 7/4
                    {
                        \grace {
                            d'16
                        }
                        ef'2
                        ~
                        ef'8
                        e'4
                        ~
                    }
                }
                {
                    \tuplet 5/4
                    {
                        e'8
                        r32
                    }
                    r2.
                }
            }
        }
        """
    ), print(string)
    assert_q_event_attachments(result, all_attachments)


def test_Quantize_13():
    definition = {"divisors": (2, 3, 5, 7), "max_depth": 2, "max_divisions": 2}
    search_tree = nauert.WeightedSearchTree(definition=definition)
    durations = [400, 1000, 1260, 840, 20, 610, 420, 2450, 3500]
    pitches = range(len(durations))
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    q_schema = nauert.MeasurewiseQSchema(
        search_tree=search_tree, time_signature=(7, 8), use_full_measure=True
    )
    result = nauert.quantize(q_event_sequence, q_schema=q_schema, attach_tempos=True)
    staff = abjad.Staff([result])
    abjad.Score([staff], name="Score")
    string = abjad.lilypond(staff)
    assert string == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \tweak text #tuplet-number::calc-fraction-text
                    \tuplet 10/7
                    {
                        \tuplet 7/4
                        {
                            \tempo 4=60
                            \time 7/8
                            c'4
                            cs'2
                            ~
                            cs'8
                        }
                        \tweak text #tuplet-number::calc-fraction-text
                        \tuplet 5/3
                        {
                            d'2.
                            ef'2
                        }
                    }
                }
                {
                    \tuplet 5/4
                    {
                        \grace {
                            e'16
                        }
                        f'8.
                        fs'8
                    }
                    g'2
                    ~
                    g'8
                }
                {
                    af'2..
                }
            }
        }
        """
    ), print(string)


def test_Quantize_14():
    definition = {"divisors": (2, 3, 5, 7), "max_depth": 2, "max_divisions": 2}
    search_tree = nauert.WeightedSearchTree(definition=definition)
    durations = [1000, 1000, 1000, 400, 50, 50, 3500]
    pitches = range(len(durations))
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    q_schema = nauert.MeasurewiseQSchema(
        search_tree=search_tree, time_signature=(7, 8), use_full_measure=True
    )
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
    result = nauert.quantize(
        q_event_sequence,
        q_schema=q_schema,
        attack_point_optimizer=attack_point_optimizer,
        attach_tempos=True,
    )
    staff = abjad.Staff([result])
    abjad.Score([staff], name="Score")
    string = abjad.lilypond(staff)
    assert string == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 7/8
                    c'4
                    cs'8
                    ~
                    cs'8
                    d'8
                    ~
                    d'8
                    ef'8
                }
                {
                    \grace {
                        e'16
                        f'16
                    }
                    fs'2..
                }
            }
        }
        """
    ), print(string)


def test_Quantize_15():
    durations = [1000, 1000, 1000, 400, 50, 50]
    pitches = range(len(durations))

    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    search_tree = nauert.UnweightedSearchTree()
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
    q_schema = nauert.MeasurewiseQSchema(
        search_tree=search_tree, time_signature=(7, 8), use_full_measure=True
    )

    result = nauert.quantize(
        q_event_sequence,
        q_schema=q_schema,
        attach_tempos=True,
        attack_point_optimizer=attack_point_optimizer,
    )

    staff = abjad.Staff([result])
    abjad.Score([staff], name="Score")
    string = abjad.lilypond(staff)
    assert string == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \tempo 4=60
                    \time 7/8
                    c'4
                    cs'8
                    ~
                    cs'8
                    d'8
                    ~
                    d'8
                    \afterGrace
                    ef'8
                    {
                        e'16
                        f'16
                    }
                }
            }
        }
        """
    ), print(string)


def test_Quantize_16():
    durations = [1546, 578, 375, 589, 144, 918, 137]
    pitches = list(range(len(durations)))
    pitches[0] = None
    all_attachments = [(x,) for x in pitches]
    all_attachments[0] = ()
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_attachment_tuples(
        tuple(zip(durations, pitches, all_attachments))
    )
    definition = {"divisors": (2, 3, 5, 7), "max_depth": 2, "max_divisions": 2}
    search_tree = nauert.WeightedSearchTree(definition=definition)
    attack_point_optimizer = nauert.MeasurewiseAttackPointOptimizer()
    q_schema = nauert.MeasurewiseQSchema(
        search_tree=search_tree,
        tempo=abjad.MetronomeMark(abjad.Duration(1, 4), 72),
        time_signature=(7, 8),
        use_full_measure=True,
    )
    result = nauert.quantize(
        q_event_sequence,
        q_schema=q_schema,
        attach_tempos=True,
        attack_point_optimizer=attack_point_optimizer,
    )
    staff = abjad.Staff([result])
    abjad.Score([staff], name="Score")
    string = abjad.lilypond(staff)
    assert string == abjad.string.normalize(
        r"""
        \new Staff
        {
            \new Voice
            {
                {
                    \tweak text #tuplet-number::calc-fraction-text
                    \tuplet 7/5
                    {
                        \tempo 4=72
                        \time 7/8
                        r4.
                        r4
                        cs'4
                    }
                    d'8
                    ef'8
                    ~
                }
                {
                    \tuplet 5/4
                    {
                        ef'16
                        e'32
                        ~
                        e'16
                    }
                    f'4
                    fs'2
                }
            }
        }
        """
    ), print(string)
    assert_q_event_attachments(result, all_attachments[1:])



import abjad

import nauert


def test_QGrid_distance_01():
    q_grid = nauert.QGrid()
    assert q_grid.distance is None

    a = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(0), ["A"]), abjad.Offset(0))
    q_grid.fit_q_events([a])
    assert q_grid.distance == abjad.Offset(0)

    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 20), ["B"]), abjad.Offset(1, 20)
    )
    q_grid.fit_q_events([b])
    assert q_grid.distance == abjad.Offset(1, 40)

    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(9, 20), ["C"]), abjad.Offset(9, 20)
    )
    q_grid.fit_q_events([c])
    assert q_grid.distance == abjad.Offset(1, 6)

    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["D"]), abjad.Offset(1, 2)
    )
    q_grid.fit_q_events([d])
    assert q_grid.distance == abjad.Offset(1, 4)

    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(11, 20), ["E"]), abjad.Offset(11, 20)
    )
    q_grid.fit_q_events([e])
    assert q_grid.distance == abjad.Offset(29, 100)

    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(19, 20), ["F"]), abjad.Offset(19, 20)
    )
    q_grid.fit_q_events([f])
    assert q_grid.distance == abjad.Offset(1, 4)

    g = nauert.QEventProxy(nauert.SilentQEvent(abjad.Offset(1), ["G"]), abjad.Offset(1))
    q_grid.fit_q_events([g])
    assert q_grid.distance == abjad.Offset(3, 14)

    q_events = q_grid.subdivide_leaves([(0, (1, 1))])
    q_grid.fit_q_events(q_events)

    assert q_grid.distance == abjad.Offset(1, 35)



import abjad

import nauert


def test_QGrid_subdivide_leaf_01():
    q_grid = nauert.QGrid()
    a = nauert.QEventProxy(nauert.PitchedQEvent(abjad.Offset(0), [0]), abjad.Offset(0))
    b = nauert.QEventProxy(
        nauert.PitchedQEvent(abjad.Offset(9, 20), [1]), abjad.Offset(9, 20)
    )
    c = nauert.QEventProxy(
        nauert.PitchedQEvent(abjad.Offset(1, 2), [2]), abjad.Offset(1, 2)
    )
    d = nauert.QEventProxy(
        nauert.PitchedQEvent(abjad.Offset(11, 20), [3]), abjad.Offset(11, 20)
    )
    e = nauert.QEventProxy(nauert.PitchedQEvent(abjad.Offset(1), [4]), abjad.Offset(1))
    q_grid.leaves[0].q_event_proxies.extend([a, b, c, d])
    q_grid.leaves[1].q_event_proxies.append(e)
    result = q_grid.subdivide_leaf(q_grid.leaves[0], (2, 3))
    assert result == [a, b, c, d]

    root_node = nauert.QGridContainer(
        (1, 1),
        children=[
            nauert.QGridLeaf(
                preprolated_duration=abjad.Duration(2, 1), q_event_proxies=[]
            ),
            nauert.QGridLeaf(
                preprolated_duration=abjad.Duration(3, 1), q_event_proxies=[]
            ),
        ],
    )
    assert format(q_grid.root_node) == format(root_node)



import abjad

import nauert


def test_QEventSequence_from_millisecond_pitch_attachment_tuples_01():
    durations = [100, 200, 100, 300, 350, 400, 600]
    pitches = [0, None, None, [1, 4], None, 5, 7]
    attachments = [("foo",), (), (), (6,), (), ("foobar",), ("foo", "bar")]
    tuples = tuple(zip(durations, pitches, attachments))
    q_events = nauert.QEventSequence.from_millisecond_pitch_attachment_tuples(tuples)
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0), (abjad.NamedPitch("c'"),), ("foo",)),
            nauert.SilentQEvent(abjad.Offset(100, 1)),
            nauert.PitchedQEvent(
                abjad.Offset(400, 1),
                (abjad.NamedPitch("cs'"), abjad.NamedPitch("e'")),
                (6,),
            ),
            nauert.SilentQEvent(abjad.Offset(700, 1)),
            nauert.PitchedQEvent(
                abjad.Offset(1050, 1), (abjad.NamedPitch("f'"),), ("foobar",)
            ),
            nauert.PitchedQEvent(
                abjad.Offset(1450, 1), (abjad.NamedPitch("g'"),), ("foo", "bar")
            ),
            nauert.TerminalQEvent(abjad.Offset(2050, 1)),
        )
    ), repr(q_events)



import abjad

import nauert

test_time_segments = [
    (116, 255),
    (279, 580),
    (627, 720),
    (743, 1300),
    (1324, 1509),
    (1533, 2090),
    (2113, 2833),
    (3320, 3390),
    (3413, 4087),
    (4203, 4389),
    (4412, 4807),
    (4946, 5898),
    (6478, 6687),
    (6711, 6780),
    (6803, 7082),
    (7129, 7314),
    (7361, 7918),
    (7988, 8290),
    (8313, 8452),
    (8475, 8731),
    (8754, 8824),
    (8847, 9009),
    (9033, 9172),
    (9404, 9474),
    (9520, 9961),
    (10356, 10449),
    (10472, 10588),
    (10612, 10960),
    (11006, 11262),
    (11285, 11378),
    (11401, 11517),
    (11540, 12237),
    (12423, 12957),
    (13073, 13166),
    (13189, 13700),
    (13769, 13862),
    (14095, 14234),
    (14350, 15279),
    (15372, 15952),
    (15999, 16091),
    (16138, 16324),
    (16765, 16997),
    (17043, 17136),
    (17160, 17345),
    (17392, 17578),
    (18599, 19342),
]


def test_QEventSequence_from_millisecond_durations_01():
    """
    Test basic functionality.
    """

    durations = abjad.math.difference_series([x[0] for x in test_time_segments])
    q_events = nauert.QEventSequence.from_millisecond_durations(durations)
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(163, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(511, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(627, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(1208, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(1417, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(1997, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(3204, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(3297, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(4087, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(4296, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(4830, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(6362, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(6595, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(6687, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(7013, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(7245, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(7872, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(8197, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(8359, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(8638, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(8731, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(8917, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(9288, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(9404, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(10240, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(10356, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(10496, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(10890, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(11169, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(11285, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(11424, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(12307, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(12957, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(13073, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(13653, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(13979, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(14234, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(15256, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(15883, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(16022, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(16649, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(16927, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(17044, 1), (abjad.NamedPitch("c'"),)),
            nauert.PitchedQEvent(abjad.Offset(17276, 1), (abjad.NamedPitch("c'"),)),
            nauert.TerminalQEvent(abjad.Offset(18483, 1)),
        )
    )


def test_QEventSequence_from_millisecond_durations_02():
    """
    Silences are not fused.
    """

    durations = [100, -100, 100, -100, -100, 100]
    q_events = nauert.QEventSequence.from_millisecond_durations(
        durations, fuse_silences=False
    )
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(100)),
            nauert.PitchedQEvent(abjad.Offset(200), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(300)),
            nauert.SilentQEvent(abjad.Offset(400)),
            nauert.PitchedQEvent(abjad.Offset(500), (abjad.NamedPitch("c'"),)),
            nauert.TerminalQEvent(abjad.Offset(600)),
        )
    )


def test_QEventSequence_from_millisecond_durations_03():
    """
    Silences are fused.
    """

    durations = [100, -100, 100, -100, -100, 100]
    q_events = nauert.QEventSequence.from_millisecond_durations(
        durations, fuse_silences=True
    )
    assert q_events == nauert.QEventSequence(
        (
            nauert.PitchedQEvent(abjad.Offset(0), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(100)),
            nauert.PitchedQEvent(abjad.Offset(200), (abjad.NamedPitch("c'"),)),
            nauert.SilentQEvent(abjad.Offset(300)),
            nauert.PitchedQEvent(abjad.Offset(500), (abjad.NamedPitch("c'"),)),
            nauert.TerminalQEvent(abjad.Offset(600)),
        )
    )



import abjad

import nauert


def test_WeightedSearchTree___call___01():
    definition = {"divisors": (2, 3, 5, 7), "max_depth": 3, "max_divisions": 2}
    search_tree = nauert.WeightedSearchTree(definition)
    q_grid = nauert.QGrid()
    a = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(0), ["A"], index=1),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 5), ["B"], index=2),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 4), ["C"], index=3),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 3), ["D"], index=4),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(2, 5), ["E"], index=5),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["F"], index=6),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    g = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(3, 5), ["G"], index=7),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    h = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(2, 3), ["H"], index=8),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    i = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(3, 4), ["I"], index=9),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    j = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(4, 5), ["J"], index=10),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    k = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1), ["K"], index=11),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    q_grid.fit_q_events([a, b, c, d, e, f, g, h, i, j, k])
    q_grids = search_tree(q_grid)
    assert [q_grid.root_node.rtm_format for q_grid in q_grids] == [
        "(1 (1 1))",
        "(1 (2 1))",
        "(1 (1 2))",
        "(1 (4 1))",
        "(1 (3 2))",
        "(1 (2 3))",
        "(1 (1 4))",
        "(1 (6 1))",
        "(1 (5 2))",
        "(1 (4 3))",
        "(1 (3 4))",
        "(1 (2 5))",
        "(1 (1 6))",
    ]



import typing

import abjad

from . import qevents as _qevents


class QEventProxy:
    """
    Q-event proxy.

    Maps Q-event offset with the range of its beatspan to the range 0-1.

    ..  container:: example

        >>> q_event = nauert.PitchedQEvent(abjad.Offset(130), [0, 1, 4])
        >>> nauert.QEventProxy(q_event, abjad.Offset(0.5))
        QEventProxy(q_event=PitchedQEvent(offset=Offset((130, 1)), pitches=...)

    Not composer-safe.

    Used internally by the ``quantize`` function.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_offset", "_q_event")

    ### INITIALIZER ###

    def __init__(
        self,
        q_event: _qevents.QEvent | None = None,
        *offsets: abjad.Offset,
    ) -> None:
        assert all(isinstance(_, abjad.Offset) for _ in offsets), repr(offsets)
        if len(offsets) == 1:
            offset = abjad.Offset(offsets[0])
            assert isinstance(q_event, _qevents.QEvent)
            assert 0 <= offset <= 1
        elif len(offsets) == 2:
            minimum, maximum = (
                abjad.Offset(offsets[0]),
                abjad.Offset(offsets[1]),
            )
            assert isinstance(q_event, _qevents.QEvent)
            assert minimum <= q_event.offset <= maximum
            offset = (q_event.offset - minimum) / (maximum - minimum)
        elif len(offsets) == 0:
            assert q_event is None
            offset = abjad.Offset(0)
        else:
            message = f"can not initialize {type(self).__name__}: {offsets!r}."
            raise ValueError(message)
        self._q_event = q_event
        self._offset = abjad.Offset(offset)

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when `argument` is a q-event proxy with offset and q-event
        equal to those of this q-event proxy. Otherwise false.
        """
        if type(self) is type(argument):
            if self.offset == argument.offset:
                if self.q_event == argument.q_event:
                    return True
        return False

    def __hash__(self) -> int:
        """
        Hashes q-event proxy.

        Required to be explicitly redefined on Python 3 if __eq__ changes.
        """
        return super(QEventProxy, self).__hash__()

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        class_name = type(self).__name__
        string = f"{class_name}(q_event={self.q_event!r}, offset={self.offset!r})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def index(self) -> typing.Optional[int]:
        """
        Gets index of q-event proxy.
        """
        assert self._q_event is not None, "There is no QEvent is this proxy."
        return self._q_event.index

    @property
    def offset(self) -> abjad.Offset:
        """
        Gets offset of q-event proxy.
        """
        return self._offset

    @property
    def q_event(self) -> typing.Optional[_qevents.QEvent]:
        """
        Gets q-event of q-event proxy.
        """
        return self._q_event



import abjad

import nauert


def test_DiscardingGraceHandler___call___01():
    grace_handler = nauert.DiscardingGraceHandler()
    durations = [1000, 1, 1, 998, 1, 999, 1, 999]
    pitches = [0, 1, 2, 3, 4, 5, 6, None]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                ef'4
                f'4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))
    assert len(grace_handler.discarded_q_events) == 3
    assert grace_handler.discarded_q_events[0] == (
        nauert.PitchedQEvent(abjad.Offset(1000), [1]),
        nauert.PitchedQEvent(abjad.Offset(1001), [2]),
    )
    assert grace_handler.discarded_q_events[1] == (
        nauert.PitchedQEvent(abjad.Offset(2000), [4]),
    )
    assert grace_handler.discarded_q_events[2] == (
        nauert.PitchedQEvent(abjad.Offset(3000), [6]),
    )



import abjad

import nauert


def test_QEventProxy___init___01():
    q_event = nauert.PitchedQEvent(abjad.Offset(130), [0])
    proxy = nauert.QEventProxy(q_event, abjad.Offset(0.5))
    assert proxy.q_event == q_event
    assert proxy.offset == abjad.Offset(1, 2)


def test_QEventProxy___init___02():
    q_event = nauert.PitchedQEvent(abjad.Offset(130), [0, 1, 4])
    proxy = nauert.QEventProxy(q_event, abjad.Offset(100), abjad.Offset(1000))
    assert proxy.q_event == q_event
    assert proxy.offset == abjad.Offset(1, 30)



import abc
import multiprocessing
import pickle
import typing

from . import quantizationjob as _quantizationjob


class JobHandler(abc.ABC):
    """
    Abstact job-handler.

    ``JobHandlers`` control how ``QuantizationJob`` instances are processed by
    the ``quantize`` function, either serially or in parallel.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self):
        pass

    ### SPECIAL METHODS ###

    @abc.abstractmethod
    def __call__(self, jobs):
        """
        Calls job handler.
        """
        raise NotImplementedError


class ParallelJobHandlerWorker(multiprocessing.Process):
    """
    Parallel job-handler worker.

    Worker process which runs ``QuantizationJobs``.

    Not composer-safe.

    Used internally by ``ParallelJobHandler``.
    """

    ### INITIALIZER ###

    def __init__(self, job_queue=None, result_queue=None) -> None:
        multiprocessing.Process.__init__(self)
        job_queue = job_queue or ()
        result_queue = result_queue or ()
        self.job_queue = job_queue
        self.result_queue = result_queue

    ### PUBLIC METHODS ###

    def run(self) -> None:
        """
        Runs parallel job handler worker.
        """
        while True:
            job = None
            if hasattr(self.job_queue, "get"):
                job = self.job_queue.get()
            if job is None:
                # poison pill causes worker shutdown
                # print '{}: Exiting'.format(process_name)
                assert hasattr(self.job_queue, "task_done")
                self.job_queue.task_done()
                break
            # print '{}: {!r}'.format(process_name, job)
            job = pickle.loads(job)
            job()
            self.job_queue.task_done()
            assert hasattr(self.result_queue, "put")
            self.result_queue.put(pickle.dumps(job, protocol=0))
        return


class ParallelJobHandler(JobHandler):
    """
    Parallel job-handler.

    Processes ``QuantizationJob`` instances in parallel, based on the number of
    CPUs available.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __call__(self, jobs):
        """
        Calls parallel job handler.
        """
        finished_jobs = []
        job_queue = multiprocessing.JoinableQueue()
        result_queue = multiprocessing.Queue()
        workers = [
            ParallelJobHandlerWorker(job_queue, result_queue)
            for i in range(multiprocessing.cpu_count() * 2)
        ]
        for worker in workers:
            worker.start()
        for job in jobs:
            job_queue.put(pickle.dumps(job, protocol=0))
        for i in range(len(jobs)):
            finished_jobs.append(pickle.loads(result_queue.get()))
        for worker in workers:
            job_queue.put(None)
        job_queue.join()
        result_queue.close()
        job_queue.close()
        for worker in workers:
            worker.join()
        return finished_jobs


class SerialJobHandler(JobHandler):
    """
    Serial job-handler.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __call__(
        self, jobs: typing.Sequence[_quantizationjob.QuantizationJob]
    ) -> typing.Sequence[_quantizationjob.QuantizationJob]:
        """
        Calls serial job handler.
        """
        for job in jobs:
            job()
        return jobs



import copy

import nauert


def test_QGrid___copy___01():
    q_grid = nauert.QGrid()
    copied = copy.deepcopy(q_grid)
    assert format(q_grid) == format(copied)
    assert q_grid != copied
    assert q_grid is not copied
    assert q_grid.root_node is not copied.root_node
    assert q_grid.next_downbeat is not copied.next_downbeat



import pickle

import abjad

import nauert


def test_QGrid_pickle_01():
    q_grid = nauert.QGrid(
        root_node=nauert.QGridContainer(
            (1, 1),
            children=[
                nauert.QGridLeaf(
                    preprolated_duration=abjad.Duration(1, 1),
                    q_event_proxies=[
                        nauert.QEventProxy(
                            nauert.SilentQEvent(abjad.Offset(100)),
                            abjad.Offset(0.5),
                        )
                    ],
                )
            ],
        ),
        next_downbeat=nauert.QGridLeaf(
            preprolated_duration=abjad.Duration(1, 1),
            q_event_proxies=[
                nauert.QEventProxy(
                    nauert.TerminalQEvent(abjad.Offset(200)), abjad.Offset(0.9)
                )
            ],
        ),
    )
    pickled = pickle.loads(pickle.dumps(q_grid))
    assert repr(pickled) == repr(q_grid)
    assert pickled is not q_grid
    assert pickled != q_grid



import abjad

import nauert


def test_CollapsingGraceHandler___call___01():
    grace_handler = nauert.CollapsingGraceHandler()
    durations = [1000, 1, 1, 997]
    pitches = [0, 7, 4, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                <c' e' g'>4
                r4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))


def test_CollapsingGraceHandler___call___02():
    grace_handler = nauert.CollapsingGraceHandler()
    durations = [1000, 1, 1, 1, 997]
    pitches = [0, 7, None, 4, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                <c' e' g'>4
                r4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))


def test_CollapsingGraceHandler___call___03():
    grace_handler = nauert.CollapsingGraceHandler()
    durations = [1000, 1, 1, 1, 997]
    pitches = [0, None, 7, 4, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                <c' e' g'>4
                r4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))


def test_CollapsingGraceHandler___call___04():
    grace_handler = nauert.CollapsingGraceHandler()
    durations = [1000, 1, 1, 1, 997]
    pitches = [0, 4, 7, None, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                <c' e' g'>4
                r4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))



import typing

from . import qeventproxy as _qeventproxy
from . import qgrid as _qgrid
from . import searchtrees as _searchtrees


class QuantizationJob:
    r"""
    Quantization job.

    Copiable, picklable class for generating all ``QGrids`` which are valid
    under a given ``SearchTree`` for a sequence of ``QEventProxies``.

    ..  container:: example

        >>> q_event_a = nauert.PitchedQEvent(abjad.Offset(250), [0, 1])
        >>> q_event_b = nauert.SilentQEvent(abjad.Offset(500))
        >>> q_event_c = nauert.PitchedQEvent(abjad.Offset(750), [3, 7])
        >>> proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.25))
        >>> proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.5))
        >>> proxy_c = nauert.QEventProxy(q_event_c, abjad.Offset(0.75))

        >>> definition = {2: {2: None}, 3: None, 5: None}
        >>> search_tree = nauert.UnweightedSearchTree(definition)

        >>> job = nauert.QuantizationJob(
        ...     1, search_tree, [proxy_a, proxy_b, proxy_c])

    ..  container:: example

        ``QuantizationJob`` generates ``QGrids`` when called, and stores those
        ``QGrids`` on its ``q_grids`` attribute, allowing them to be recalled
        later, even if pickled:

        >>> job()
        >>> for q_grid in job.q_grids:
        ...     print(q_grid.rtm_format)
        1
        (1 (1 1 1 1 1))
        (1 (1 1 1))
        (1 (1 1))
        (1 ((1 (1 1)) (1 (1 1))))

    ``QuantizationJob`` is intended to be useful in multiprocessing-enabled
    environments.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_job_id", "_q_event_proxies", "_q_grids", "_search_tree")

    ### INITIALIZER ###

    def __init__(
        self,
        job_id: int = 1,
        search_tree: _searchtrees.SearchTree | None = None,
        q_event_proxies: typing.Sequence[_qeventproxy.QEventProxy] | None = None,
        q_grids: typing.Sequence[_qgrid.QGrid] | None = None,
    ):
        search_tree = search_tree or _searchtrees.UnweightedSearchTree()
        q_event_proxies = q_event_proxies or []
        assert isinstance(search_tree, _searchtrees.SearchTree)
        assert all(isinstance(x, _qeventproxy.QEventProxy) for x in q_event_proxies)
        self._job_id = job_id
        self._search_tree = search_tree
        self._q_event_proxies = tuple(q_event_proxies)
        self._q_grids: tuple[_qgrid.QGrid, ...]
        if q_grids is None:
            self._q_grids = ()
        else:
            assert all(isinstance(x, _qgrid.QGrid) for x in q_grids)
            self._q_grids = tuple(q_grids)

    ### SPECIAL METHODS ###

    def __call__(self) -> None:
        """
        Calls quantization job.
        """
        # print('XXX')
        # print(format(self.q_event_proxies[0]))
        q_grid = _qgrid.QGrid()
        q_grid.fit_q_events(self.q_event_proxies)
        # print(format(q_grid))
        old_q_grids = []
        new_q_grids = [q_grid]
        while new_q_grids:
            q_grid = new_q_grids.pop()
            search_results = self.search_tree(q_grid)
            # print q_grid.rtm_format
            # for x in search_results:
            #    print '\t', x.rtm_format
            new_q_grids.extend(search_results)
            old_q_grids.append(q_grid)
        # for q_grid in old_q_grids:
        #    print('\t', q_grid)
        # print()

        self._q_grids = tuple(old_q_grids)

    def __eq__(self, argument) -> bool:
        """
        Is true when `argument` is a quantization job with job ID, search tree,
        q-event proxies and q-grids equal to those of this quantization job.
        Otherwise false.
        """
        if type(self) is type(argument):
            if self.job_id == argument.job_id:
                if self.search_tree == argument.search_tree:
                    if self.q_event_proxies == argument.q_event_proxies:
                        if self.q_grids == argument.q_grids:
                            return True
        return False

    def __hash__(self) -> int:
        """
        Hashes quantization job.

        Required to be explicitly redefined on Python 3 if __eq__ changes.
        """
        return super(QuantizationJob, self).__hash__()

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(job_id={self.job_id!r},"
        string += f" search_tree={self.search_tree!r},"
        string += f" q_event_proxies={self.q_event_proxies!r}, q_grids={self.q_grids})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def job_id(self) -> int:
        """
        Gets the job id of the ``QuantizationJob``.

        Only meaningful when the job is processed via multiprocessing, as the
        job id is necessary to reconstruct the order of jobs.
        """
        return self._job_id

    @property
    def q_event_proxies(self) -> tuple[_qeventproxy.QEventProxy, ...]:
        r"""
        Gets the ``QEventProxies`` the ``QuantizationJob`` was instantiated with.

        >>> q_event_a = nauert.PitchedQEvent(abjad.Offset(250), [0, 1])
        >>> q_event_b = nauert.SilentQEvent(abjad.Offset(500))
        >>> q_event_c = nauert.PitchedQEvent(abjad.Offset(750), [3, 7])
        >>> proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.25))
        >>> proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.5))
        >>> proxy_c = nauert.QEventProxy(q_event_c, abjad.Offset(0.75))

        >>> definition = {2: {2: None}, 3: None, 5: None}
        >>> search_tree = nauert.UnweightedSearchTree(definition)

        >>> job = nauert.QuantizationJob(
        ...     1, search_tree, [proxy_a, proxy_b, proxy_c])
        >>> job()

        >>> for q_event_proxy in job.q_event_proxies:
        ...     print(type(q_event_proxy.q_event).__name__)
        ...
        PitchedQEvent
        SilentQEvent
        PitchedQEvent

        """
        return self._q_event_proxies

    @property
    def q_grids(self) -> tuple[_qgrid.QGrid, ...]:
        r"""
        Gets generated ``QGrids``.

        >>> q_event_a = nauert.PitchedQEvent(abjad.Offset(250), [0, 1])
        >>> q_event_b = nauert.SilentQEvent(abjad.Offset(500))
        >>> q_event_c = nauert.PitchedQEvent(abjad.Offset(750), [3, 7])
        >>> proxy_a = nauert.QEventProxy(q_event_a, abjad.Offset(0.25))
        >>> proxy_b = nauert.QEventProxy(q_event_b, abjad.Offset(0.5))
        >>> proxy_c = nauert.QEventProxy(q_event_c, abjad.Offset(0.75))
        >>> definition = {2: {2: None}, 3: None, 5: None}
        >>> search_tree = nauert.UnweightedSearchTree(definition)
        >>> job = nauert.QuantizationJob(1, search_tree, [proxy_a, proxy_b, proxy_c])
        >>> job()

        >>> for q_grid in job.q_grids:
        ...     print(q_grid.rtm_format)
        1
        (1 (1 1 1 1 1))
        (1 (1 1 1))
        (1 (1 1))
        (1 ((1 (1 1)) (1 (1 1))))

        """
        return self._q_grids

    @property
    def search_tree(self) -> _searchtrees.SearchTree:
        """
        Gets search tree ``QuantizationJob`` was instantiated with.
        """
        return self._search_tree



import abjad

import nauert


def test_MeasurewiseQSchema___init___01():
    item_a = nauert.MeasurewiseQSchemaItem(
        search_tree=nauert.UnweightedSearchTree({2: None})
    )
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 76)
    item_b = nauert.MeasurewiseQSchemaItem(tempo=metronome_mark)
    time_signature = abjad.TimeSignature((3, 4))
    item_c = nauert.MeasurewiseQSchemaItem(time_signature=time_signature)
    item_d = nauert.MeasurewiseQSchemaItem(use_full_measure=True)
    time_signature = abjad.TimeSignature((5, 8))
    schema = nauert.MeasurewiseQSchema(
        {2: item_a, 4: item_b, 7: item_c, 8: item_d},
        search_tree=nauert.UnweightedSearchTree({3: None}),
        tempo=(abjad.Duration(1, 8), 58),
        time_signature=time_signature,
        use_full_measure=False,
    )
    assert len(schema.items) == 4
    assert schema.search_tree == nauert.UnweightedSearchTree({3: None})
    assert schema.tempo == abjad.MetronomeMark(abjad.Duration(1, 8), 58)
    assert schema.time_signature == abjad.TimeSignature((5, 8))
    assert schema.use_full_measure is False


def test_MeasurewiseQSchema___init___02():
    schema = nauert.MeasurewiseQSchema()
    assert len(schema.items) == 0
    assert schema.search_tree == nauert.UnweightedSearchTree()
    assert schema.tempo == abjad.MetronomeMark(abjad.Duration(1, 4), 60)
    assert schema.time_signature == abjad.TimeSignature((4, 4))
    assert schema.use_full_measure is False



import nauert


def test_UnweightedSearchTree___init___01():
    search_tree = nauert.UnweightedSearchTree()
    assert search_tree.definition == search_tree.default_definition


def test_UnweightedSearchTree___init___02():
    definition = {2: None, 3: {2: None}}
    search_tree = nauert.UnweightedSearchTree(definition)
    assert search_tree.definition == definition



import abc
import bisect
import copy

import abjad

from . import qschemaitems as _qschemaitems
from . import qtargetitems as _qtargetitems
from . import qtargets as _qtargets
from . import searchtrees as _searchtrees


class QSchema(abc.ABC):
    """
    Abstract Q-schema.

    ``QSchema`` allows for the specification of quantization settings
    diachronically, at any time-step of the quantization process.

    In practice, this provides a means for the composer to change the tempo,
    search-tree, time-signature etc., effectively creating a template into
    which quantized rhythms can be "poured", without yet knowing what those
    rhythms might be, or even how much time the ultimate result will take. Like
    Abjad indicators the settings made at any given time-step via a ``QSchema``
    instance are understood to persist until changed.

    All concrete ``QSchema`` subclasses strongly implement default values for
    all of their parameters.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_items", "_lookups")

    _keyword_argument_names: tuple[str, ...] = ()

    _search_tree = _searchtrees.UnweightedSearchTree()

    _tempo = abjad.MetronomeMark()

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, *arguments, **keywords):
        if 1 == len(arguments) and isinstance(arguments[0], type(self)):
            items = copy.deepcopy(arguments[0].items)
        elif 1 == len(arguments) and isinstance(arguments[0], dict):
            items = list(arguments[0].items())
            if abjad.math.all_are_pairs_of_types(items, int, dict):
                items = [(x, self.item_class(**y)) for x, y in items]
            assert abjad.math.all_are_pairs_of_types(items, int, self.item_class)
            items = dict(items)
        elif abjad.math.all_are_pairs_of_types(arguments, int, self.item_class):
            items = dict(arguments)
        elif abjad.math.all_are_pairs_of_types(arguments, int, dict):
            items = [(x, self.item_class(**y)) for x, y in arguments]
            items = dict(items)
        elif all(isinstance(x, self.item_class) for x in arguments):
            items = [(i, x) for i, x in enumerate(arguments)]
            items = dict(items)
        elif all(isinstance(x, dict) for x in arguments):
            items = [(i, self.item_class(**x)) for i, x in enumerate(arguments)]
            items = dict(items)
        else:
            raise ValueError
        if items:
            assert 0 <= min(items)
        self._items = dict(items)
        self._lookups = self._create_lookups()

    ### SPECIAL METHODS ###

    def __call__(self, duration: abjad.Duration) -> _qtargets.QTarget:
        """
        Calls QSchema on ``duration``.
        """
        assert isinstance(duration, abjad.Duration), repr(duration)
        target_items = []
        idx, current_offset = 0, abjad.Offset(0)
        duration = abjad.Duration(duration)
        while current_offset < duration:
            lookup = self[idx]
            lookup["offset_in_ms"] = current_offset
            target_item = self.target_item_class(**lookup)
            target_items.append(target_item)
            current_offset += target_item.duration_in_ms
            idx += 1
        return self.target_class(target_items)

    def __getitem__(self, argument: int) -> dict:
        """
        Gets item or slice identified by `argument`.
        """
        assert isinstance(argument, int) and 0 <= argument
        result = {}
        for field in self._lookups:
            lookup = self._lookups[field].get(argument)
            if lookup is not None:
                result[field] = lookup
            else:
                keys = sorted(self._lookups[field].keys())
                idx = bisect.bisect(keys, argument)
                if len(keys) == idx:
                    key = keys[-1]
                elif argument < keys[idx]:
                    key = keys[idx - 1]
                result[field] = self._lookups[field][key]
        return result

    ### PRIVATE METHODS ###

    def _create_lookups(self) -> dict[str, dict]:
        names = self._keyword_argument_names
        lookups = {}
        for name in names:
            lookups[name] = {0: getattr(self, name)}
            for position, item in self.items.items():
                value = getattr(item, name)
                if value is not None:
                    lookups[name][position] = value
            lookups[name] = dict(lookups[name])
        return dict(lookups)

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def item_class(self):
        """
        Gets schema's item class.
        """
        raise NotImplementedError

    @property
    def items(self) -> dict[int, _qschemaitems.QSchemaItem]:
        """
        Gets items dictionary.
        """
        return self._items

    @property
    def search_tree(self) -> _searchtrees.SearchTree:
        """
        Gets default search tree.
        """
        return self._search_tree

    @abc.abstractproperty
    def target_class(self):
        """
        Gets schema's target class.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def target_item_class(self):
        """
        Gets schema's target class' item class.
        """
        raise NotImplementedError

    @property
    def tempo(self) -> abjad.MetronomeMark:
        """
        Gets default tempo.
        """
        return self._tempo


class BeatwiseQSchema(QSchema):
    r"""
    Beatwise q-schema.

    Treats beats as timestep unit.

        >>> q_schema = nauert.BeatwiseQSchema()

        Without arguments, it uses smart defaults:

    ..  container:: example

        Each time-step in a ``BeatwiseQSchema`` is composed of three settings:

            * ``beatspan``
            * ``search_tree``
            * ``tempo``

        These settings can be applied as global defaults for the schema via keyword
        arguments, which persist until overridden:

        >>> beatspan = abjad.Duration(5, 16)
        >>> search_tree = nauert.UnweightedSearchTree({7: None})
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 54)
        >>> q_schema = nauert.BeatwiseQSchema(
        ...     beatspan=beatspan,
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ... )

    ..  container:: example

        The computed value at any non-negative time-step can be found by
        subscripting:

        >>> index = 0
        >>> for key, value in sorted(q_schema[index].items()):
        ...     print("{}:".format(key), value)
        ...
        beatspan: 5/16
        search_tree: UnweightedSearchTree(definition={7: None})
        tempo: MetronomeMark(...)

        >>> index = 1000
        >>> for key, value in sorted(q_schema[index].items()):
        ...     print("{}:".format(key), value)
        ...
        beatspan: 5/16
        search_tree: UnweightedSearchTree(definition={7: None})
        tempo: MetronomeMark(...)

    ..  container:: example

        Per-time-step settings can be applied in a variety of ways.

        Instantiating the schema via ``*arguments`` with a series of either
        ``BeatwiseQSchemaItem`` instances, or dictionaries which could be used to
        instantiate ``BeatwiseQSchemaItem`` instances, will apply those settings
        sequentially, starting from time-step ``0``:

        >>> a = {"beatspan": abjad.Duration(5, 32)}
        >>> b = {"beatspan": abjad.Duration(3, 16)}
        >>> c = {"beatspan": abjad.Duration(1, 8)}

        >>> q_schema = nauert.BeatwiseQSchema(a, b, c)

        >>> q_schema[0]["beatspan"]
        Duration(5, 32)

        >>> q_schema[1]["beatspan"]
        Duration(3, 16)

        >>> q_schema[2]["beatspan"]
        Duration(1, 8)

        >>> q_schema[3]["beatspan"]
        Duration(1, 8)

    ..  container:: example

        Similarly, instantiating the schema from a single dictionary, consisting
        of integer:specification pairs, or a sequence via ``*arguments`` of (integer,
        specification) pairs, allows for applying settings to  non-sequential
        time-steps:

        >>> a = {"search_tree": nauert.UnweightedSearchTree({2: None})}
        >>> b = {"search_tree": nauert.UnweightedSearchTree({3: None})}

        >>> settings = {
        ...     2: a,
        ...     4: b,
        ... }

        >>> q_schema = nauert.BeatwiseQSchema(settings)

        >>> import pprint
        >>> ust = q_schema[0]["search_tree"]
        >>> pprint.pprint(ust.definition)
        {2: {2: {2: {2: None}, 3: None}, 3: None, 5: None, 7: None},
        3: {2: {2: None}, 3: None, 5: None},
        5: {2: None, 3: None},
        7: {2: None},
        11: None,
        13: None}

        >>> ust = q_schema[1]["search_tree"]
        >>> pprint.pprint(ust.definition)
        {2: {2: {2: {2: None}, 3: None}, 3: None, 5: None, 7: None},
        3: {2: {2: None}, 3: None, 5: None},
        5: {2: None, 3: None},
        7: {2: None},
        11: None,
        13: None}

        >>> q_schema[2]["search_tree"]
        UnweightedSearchTree(definition={2: None})

        >>> q_schema[3]["search_tree"]
        UnweightedSearchTree(definition={2: None})

        >>> q_schema[4]["search_tree"]
        UnweightedSearchTree(definition={3: None})

        >>> q_schema[1000]["search_tree"]
        UnweightedSearchTree(definition={3: None})

    ..  container:: example

        The following is equivalent to the above schema definition:

        >>> q_schema = nauert.BeatwiseQSchema(
        ...     (2, {"search_tree": nauert.UnweightedSearchTree({2: None})}),
        ...     (4, {"search_tree": nauert.UnweightedSearchTree({3: None})}),
        ... )

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_beatspan", "_items", "_lookups", "_search_tree", "_tempo")

    _keyword_argument_names = ("beatspan", "search_tree", "tempo")

    ### INITIALIZER ###

    def __init__(self, *arguments, **keywords):
        self._beatspan = abjad.Duration(keywords.get("beatspan", (1, 4)))
        search_tree = keywords.get("search_tree", _searchtrees.UnweightedSearchTree())
        assert isinstance(search_tree, _searchtrees.SearchTree)
        self._search_tree = search_tree
        tempo = keywords.get("tempo", (abjad.Duration(1, 4), 60))
        if isinstance(tempo, tuple):
            tempo = abjad.MetronomeMark(*tempo)
        self._tempo = tempo
        QSchema.__init__(self, *arguments, **keywords)

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(beatspan={self.beatspan!r},"
        string += f" search_tree={self.search_tree!r}, tempo={self.tempo!r})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def beatspan(self) -> abjad.Duration:
        """
        Gets default beatspan of beatwise q-schema.
        """
        return self._beatspan

    @property
    def item_class(self) -> type[_qschemaitems.BeatwiseQSchemaItem]:
        """
        Gets schema's item class.
        """
        return _qschemaitems.BeatwiseQSchemaItem

    @property
    def target_class(self) -> type[_qtargets.BeatwiseQTarget]:
        """
        Gets target class of beatwise q-schema.
        """
        return _qtargets.BeatwiseQTarget

    @property
    def target_item_class(self) -> type[_qtargetitems.QTargetBeat]:
        """
        Gets target item class of beatwise q-schema.
        """
        return _qtargetitems.QTargetBeat


class MeasurewiseQSchema(QSchema):
    r"""
    Measurewise q-schema.

    Treats measures as its timestep unit.

    >>> q_schema = nauert.MeasurewiseQSchema()

    ..  container:: example

        Without arguments, it uses smart defaults:

        >>> import pprint
        >>> pprint.pprint(q_schema.search_tree.definition)
        {2: {2: {2: {2: None}, 3: None}, 3: None, 5: None, 7: None},
        3: {2: {2: None}, 3: None, 5: None},
        5: {2: None, 3: None},
        7: {2: None},
        11: None,
        13: None}

    ..  container:: example

        Each time-step in a ``MeasurewiseQSchema`` is composed of four settings:

            * ``search_tree``
            * ``tempo``
            * ``time_signature``
            * ``use_full_measure``

        These settings can be applied as global defaults for the schema via keyword
        arguments, which persist until overridden:

        >>> search_tree = nauert.UnweightedSearchTree({7: None})
        >>> time_signature = abjad.TimeSignature((3, 4))
        >>> tempo = abjad.MetronomeMark(abjad.Duration(1, 4), 54)
        >>> use_full_measure = True
        >>> q_schema = nauert.MeasurewiseQSchema(
        ...     search_tree=search_tree,
        ...     tempo=tempo,
        ...     time_signature=time_signature,
        ...     use_full_measure=use_full_measure,
        ... )

        All of these settings are self-descriptive, except for
        ``use_full_measure``, which controls whether the measure is subdivided
        by the ``quantize`` function into beats according to its time
        signature.

        If ``use_full_measure`` is ``False``, the time-step's measure will be
        divided into units according to its time-signature.  For example, a 4/4
        measure will be divided into 4 units, each having a beatspan of 1/4.

        On the other hand, if ``use_full_measure`` is set to ``True``, the
        time-step's measure will not be subdivided into independent
        quantization units. This usually results in full-measure tuplets.

    ..  container:: example

        The computed value at any non-negative time-step can be found by
        subscripting:

        >>> index = 0
        >>> for key, value in sorted(q_schema[index].items()):
        ...     print("{}:".format(key), value)
        ...
        search_tree: UnweightedSearchTree(definition={7: None})
        tempo: MetronomeMark(...)
        time_signature: TimeSignature(pair=(3, 4), hide=False, partial=None)
        use_full_measure: True

        >>> index = 1000
        >>> for key, value in sorted(q_schema[index].items()):
        ...     print("{}:".format(key), value)
        ...
        search_tree: UnweightedSearchTree(definition={7: None})
        tempo: MetronomeMark(...)
        time_signature: TimeSignature(pair=(3, 4), hide=False, partial=None)
        use_full_measure: True

    ..  container:: example

        Per-time-step settings can be applied in a variety of ways.

        Instantiating the schema via ``*arguments`` with a series of either
        ``MeasurewiseQSchemaItem`` instances, or dictionaries which could be
        used to instantiate ``MeasurewiseQSchemaItem`` instances, will apply
        those settings sequentially, starting from time-step ``0``:

        >>> a = {"search_tree": nauert.UnweightedSearchTree({2: None})}
        >>> b = {"search_tree": nauert.UnweightedSearchTree({3: None})}
        >>> c = {"search_tree": nauert.UnweightedSearchTree({5: None})}

        >>> q_schema = nauert.MeasurewiseQSchema(a, b, c)

        >>> q_schema[0]["search_tree"]
        UnweightedSearchTree(definition={2: None})

        >>> q_schema[1]["search_tree"]
        UnweightedSearchTree(definition={3: None})

        >>> q_schema[2]["search_tree"]
        UnweightedSearchTree(definition={5: None})

        >>> q_schema[1000]["search_tree"]
        UnweightedSearchTree(definition={5: None})

    ..  container:: example

        Similarly, instantiating the schema from a single dictionary,
        consisting of integer:specification pairs, or a sequence via
        ``*arguments`` of (integer, specification) pairs, allows for applying
        settings to non-sequential time-steps:

        >>> a = {"time_signature": abjad.TimeSignature((7, 32))}
        >>> b = {"time_signature": abjad.TimeSignature((3, 4))}
        >>> c = {"time_signature": abjad.TimeSignature((5, 8))}

        >>> settings = {
        ...     2: a,
        ...     4: b,
        ...     6: c,
        ... }

        >>> q_schema = nauert.MeasurewiseQSchema(settings)

        >>> q_schema[0]["time_signature"]
        TimeSignature(pair=(4, 4), hide=False, partial=None)

        >>> q_schema[1]["time_signature"]
        TimeSignature(pair=(4, 4), hide=False, partial=None)

        >>> q_schema[2]["time_signature"]
        TimeSignature(pair=(7, 32), hide=False, partial=None)

        >>> q_schema[3]["time_signature"]
        TimeSignature(pair=(7, 32), hide=False, partial=None)

        >>> q_schema[4]["time_signature"]
        TimeSignature(pair=(3, 4), hide=False, partial=None)

        >>> q_schema[5]["time_signature"]
        TimeSignature(pair=(3, 4), hide=False, partial=None)

        >>> q_schema[6]["time_signature"]
        TimeSignature(pair=(5, 8), hide=False, partial=None)

        >>> q_schema[1000]["time_signature"]
        TimeSignature(pair=(5, 8), hide=False, partial=None)

    ..  container:: example

        The following is equivalent to the above schema definition:

        >>> q_schema = nauert.MeasurewiseQSchema(
        ...     (2, {"time_signature": abjad.TimeSignature((7, 32))}),
        ...     (4, {"time_signature": abjad.TimeSignature((3, 4))}),
        ...     (6, {"time_signature": abjad.TimeSignature((5, 8))}),
        ... )

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_items",
        "_lookups",
        "_search_tree",
        "_tempo",
        "_time_signature",
        "_use_full_measure",
    )

    _keyword_argument_names = (
        "search_tree",
        "tempo",
        "time_signature",
        "use_full_measure",
    )

    ### INITIALIZER ###

    def __init__(self, *arguments, **keywords):
        search_tree = keywords.get("search_tree", _searchtrees.UnweightedSearchTree())
        assert isinstance(search_tree, _searchtrees.SearchTree)
        self._search_tree = search_tree
        tempo = keywords.get("tempo", (abjad.Duration(1, 4), 60))
        if isinstance(tempo, tuple):
            tempo = abjad.MetronomeMark(*tempo)
        self._tempo = tempo
        time_signature = keywords.get("time_signature", (4, 4))
        if isinstance(time_signature, abjad.TimeSignature):
            self._time_signature = abjad.TimeSignature(time_signature.pair)
        elif isinstance(time_signature, tuple):
            self._time_signature = abjad.TimeSignature(time_signature)
        else:
            raise TypeError(time_signature)
        self._use_full_measure = bool(keywords.get("use_full_measure"))
        QSchema.__init__(self, *arguments, **keywords)

    def __repr__(self) -> str:
        """
        Gets repr.
        """
        string = f"{type(self).__name__}(search_tree={self.search_tree!r},"
        string += f" tempo={self.tempo!r}, time_signature={self.time_signature!r},"
        string += f" use_full_measure={self.use_full_measure})"
        return string

    ### PUBLIC PROPERTIES ###

    @property
    def item_class(self) -> type[_qschemaitems.MeasurewiseQSchemaItem]:
        """
        Gets item class of measurewise q-schema.
        """
        return _qschemaitems.MeasurewiseQSchemaItem

    @property
    def target_class(self) -> type[_qtargets.MeasurewiseQTarget]:
        """
        Gets target class of measurewise q-schema.
        """
        return _qtargets.MeasurewiseQTarget

    @property
    def target_item_class(self) -> type[_qtargetitems.QTargetMeasure]:
        """
        Gets target item class of measurewise q-schema.
        """
        return _qtargetitems.QTargetMeasure

    @property
    def time_signature(self) -> abjad.TimeSignature:
        """
        Gets default time signature of measurewise q-schema.

        ..  container:: example

            >>> q_schema = nauert.MeasurewiseQSchema(
            ...     time_signature=abjad.TimeSignature((3, 4))
            ... )
            >>> q_schema.time_signature
            TimeSignature(pair=(3, 4), hide=False, partial=None)

        ..  container:: example

            If there are multiple time signatures in the QSchema, this returns
            the default time signature of (4, 4).

            >>> a = {"time_signature": abjad.TimeSignature((7, 32))}
            >>> b = {"time_signature": abjad.TimeSignature((3, 4))}
            >>> c = {"time_signature": abjad.TimeSignature((5, 8))}

            >>> settings = {
            ...     2: a,
            ...     4: b,
            ...     6: c,
            ... }

            >>> q_schema = nauert.MeasurewiseQSchema(settings)
            >>> q_schema.time_signature
            TimeSignature(pair=(4, 4), hide=False, partial=None)

        """
        return self._time_signature

    @property
    def use_full_measure(self) -> bool:
        """
        Is true when class uses full-measure-as-beatspan default.
        """
        return self._use_full_measure



import abjad

import nauert


def test_QGridLeaf___eq___01():
    a = nauert.QGridLeaf(abjad.Duration(1), [])
    b = nauert.QGridLeaf(abjad.Duration(1), [])
    assert format(a) == format(b)
    assert a != b


def test_QGridLeaf___eq___02():
    a = nauert.QGridLeaf(abjad.Duration(1), [])
    sqe = nauert.SilentQEvent(abjad.Offset(1000))
    b = nauert.QGridLeaf(
        abjad.Duration(1),
        [nauert.QEventProxy(sqe, abjad.Offset(0.5))],
    )
    c = nauert.QGridLeaf(abjad.Duration(2), [])
    d = nauert.QGridLeaf(
        abjad.Duration(2),
        [nauert.QEventProxy(sqe, abjad.Offset(0.5))],
    )
    assert a != b
    assert a != c
    assert a != d
    assert b != c
    assert b != d
    assert c != d



import abjad
import pytest

import nauert


class Job:

    def __init__(self, number):
        self.number = number

    def __call__(self):
        self.result = [
            x for x in abjad.math.yield_all_compositions_of_integer(self.number)
        ]


@pytest.mark.skip()
def test_ParallelJobHandler___call___01():
    jobs = [Job(x) for x in range(1, 11)]
    job_handler = nauert.ParallelJobHandler()
    job_handler(jobs)


@pytest.mark.skip()
def test_ParallelJobHandler___call___02():
    job_id = 1
    definition = {2: {2: {2: None}, 3: None}, 5: None}
    search_tree = nauert.UnweightedSearchTree(definition)
    q_event_proxies = [
        nauert.QEventProxy(nauert.SilentQEvent(0, ["A"], index=1), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((1, 5), ["B"], index=2), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((1, 4), ["C"], index=3), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((1, 3), ["D"], index=4), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((2, 5), ["E"], index=5), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((1, 2), ["F"], index=6), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((3, 5), ["G"], index=7), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((2, 3), ["H"], index=8), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((3, 4), ["I"], index=9), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent((4, 5), ["J"], index=10), 0, 1),
        nauert.QEventProxy(nauert.SilentQEvent(1, ["K"], index=11), 0, 1),
    ]
    job_a = nauert.QuantizationJob(job_id, search_tree, q_event_proxies)
    job_b = nauert.QuantizationJob(job_id, search_tree, q_event_proxies)
    assert job_a == job_b

    a_jobs = nauert.SerialJobHandler()([job_a])
    b_jobs = nauert.ParallelJobHandler()([job_b])
    assert len(a_jobs) == len(b_jobs)

    a_rtms = sorted([q_grid.root_node.rtm_format for q_grid in a_jobs[0].q_grids])
    b_rtms = sorted([q_grid.root_node.rtm_format for q_grid in b_jobs[0].q_grids])
    assert a_rtms == b_rtms
    assert sorted(a_jobs[0].q_grids, key=lambda x: x.root_node.rtm_format) == sorted(
        b_jobs[0].q_grids, key=lambda x: x.root_node.rtm_format
    )



import abjad

import nauert


def test_ConcatenatingGraceHandler___call___01():
    grace_handler = nauert.ConcatenatingGraceHandler()
    durations = [1000, 1, 999]
    pitches = [0, None, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                c'4
                r4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))


def test_ConcatenatingGraceHandler___call___02():
    grace_handler = nauert.ConcatenatingGraceHandler(discard_grace_rest=False)
    durations = [1000, 1, 999]
    pitches = [0, None, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                \grace {
                    r16
                }
                c'4
                r4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))


def test_ConcatenatingGraceHandler___call___03():
    grace_handler = nauert.ConcatenatingGraceHandler(
        replace_rest_with_final_grace_note=False
    )
    durations = [1000, 1, 999, 1000]
    pitches = [0, 0, None, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                \grace {
                    c'16
                }
                r4
                c'4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))


def test_ConcatenatingGraceHandler___call___04():
    grace_handler = nauert.ConcatenatingGraceHandler(
        replace_rest_with_final_grace_note=True
    )
    durations = [1000, 1, 999, 1000]
    pitches = [0, 0, None, 0]
    q_event_sequence = nauert.QEventSequence.from_millisecond_pitch_pairs(
        tuple(zip(durations, pitches))
    )
    result = nauert.quantize(q_event_sequence, grace_handler=grace_handler)
    assert abjad.lilypond(result) == abjad.string.normalize(
        r"""
        \new Voice
        {
            {
                %%% \time 4/4 %%%
                \tempo 4=60
                c'4
                c'4
                c'4
                r4
            }
        }
        """
    ), print(abjad.lilypond(result))



import abjad

import nauert


def test_QGridContainer___eq___01():
    a = nauert.QGridContainer((1, 1), children=[])
    b = nauert.QGridContainer((1, 1), children=[])
    assert format(a) == format(b)
    assert a != b


def test_QGridContainer___eq___02():
    a = nauert.QGridContainer(
        (1, 1),
        children=[nauert.QGridLeaf(preprolated_duration=abjad.Duration(1, 1))],
    )
    b = nauert.QGridContainer(
        (1, 1),
        children=[nauert.QGridLeaf(preprolated_duration=abjad.Duration(1, 1))],
    )
    assert format(a) == format(b)
    assert a != b


def test_QGridContainer___eq___03():
    a = nauert.QGridContainer((1, 1), children=[])
    b = nauert.QGridContainer((2, 1), children=[])
    c = nauert.QGridContainer(
        (1, 1),
        children=[nauert.QGridLeaf(preprolated_duration=abjad.Duration(1, 1))],
    )
    d = nauert.QGridContainer(
        (2, 1),
        children=[nauert.QGridLeaf(preprolated_duration=abjad.Duration(1, 1))],
    )
    e = nauert.QGridContainer(
        (2, 1),
        children=[nauert.QGridLeaf(abjad.Duration(2))],
    )
    assert a != b
    assert a != c
    assert a != d
    assert a != e
    assert b != c
    assert b != d
    assert b != e
    assert c != d
    assert c != e
    assert d != e



import copy

import abjad

import nauert


def test_UnweightedSearchTree__generate_all_subdivision_commands_01():
    definition = {2: {2: {2: None}, 3: None}, 5: None}
    search_tree = nauert.UnweightedSearchTree(definition)
    q_grid = nauert.QGrid()
    a = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(0), ["A"], index=1),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    b = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 5), ["B"], index=2),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    c = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 4), ["C"], index=3),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    d = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 3), ["D"], index=4),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    e = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(2, 5), ["E"], index=5),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    f = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1, 2), ["F"], index=6),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    g = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(3, 5), ["G"], index=7),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    h = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(2, 3), ["H"], index=8),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    i = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(3, 4), ["I"], index=9),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    j = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(4, 5), ["J"], index=10),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    k = nauert.QEventProxy(
        nauert.SilentQEvent(abjad.Offset(1), ["K"], index=11),
        abjad.Offset(0),
        abjad.Offset(1),
    )
    q_grid.fit_q_events([a, b, c, d, e, f, g, h, i, j, k])
    commands = search_tree._generate_all_subdivision_commands(q_grid)
    assert commands == (((0, (1, 1)),), ((0, (1, 1, 1, 1, 1)),))

    new_q_grid = copy.deepcopy(q_grid)
    q_event_proxies = new_q_grid.subdivide_leaves(commands[0])
    new_q_grid.fit_q_events(q_event_proxies)
    new_q_grid.sort_q_events_by_index()
    new_commands = search_tree._generate_all_subdivision_commands(new_q_grid)
    assert new_q_grid.leaves[0].q_event_proxies == [a, b, c]
    assert new_q_grid.leaves[1].q_event_proxies == [d, e, f, g, h, i]
    assert new_q_grid.leaves[2].q_event_proxies == [j, k]
    assert new_commands == (
        ((0, (1, 1)), (1, (1, 1))),
        ((0, (1, 1)), (1, (1, 1, 1))),
        ((0, (1, 1, 1)), (1, (1, 1))),
        ((0, (1, 1, 1)), (1, (1, 1, 1))),
    )

    new_q_grid = copy.deepcopy(q_grid)
    q_event_proxies = new_q_grid.subdivide_leaves(commands[1])
    new_q_grid.fit_q_events(q_event_proxies)
    new_q_grid.sort_q_events_by_index()
    new_commands = search_tree._generate_all_subdivision_commands(new_q_grid)
    assert new_q_grid.leaves[0].q_event_proxies == [a]
    assert new_q_grid.leaves[1].q_event_proxies == [b, c]
    assert new_q_grid.leaves[2].q_event_proxies == [d, e, f]
    assert new_q_grid.leaves[3].q_event_proxies == [g, h]
    assert new_q_grid.leaves[4].q_event_proxies == [i, j]
    assert new_q_grid.leaves[5].q_event_proxies == [k]
    assert new_commands == ()



import decimal

import abjad
import pytest

import nauert


def test_PitchedQEvent___init___01():
    q_event = nauert.PitchedQEvent(abjad.Offset(130), [0, 1, 4])
    assert q_event.offset == abjad.Offset(130)
    assert q_event.pitches == (
        abjad.NamedPitch(0),
        abjad.NamedPitch(1),
        abjad.NamedPitch(4),
    )
    assert q_event.attachments == ()


def test_PitchedQEvent___init___02():
    q_event = nauert.PitchedQEvent(
        abjad.Offset(133, 5),
        [abjad.NamedPitch("fss")],
        attachments=["foo", "bar", "baz"],
    )
    assert q_event.offset == abjad.Offset(133, 5)
    assert q_event.pitches == (abjad.NamedPitch("fss"),)
    assert q_event.attachments == ("foo", "bar", "baz")


def test_PitchedQEvent___eq___01():
    a = nauert.PitchedQEvent(abjad.Offset(1000), [0])
    b = nauert.PitchedQEvent(abjad.Offset(1000), [0])
    assert a == b


def test_PitchedQEvent___eq___02():
    a = nauert.PitchedQEvent(abjad.Offset(1000), [0])
    b = nauert.PitchedQEvent(abjad.Offset(1000), [0], ["foo", "bar", "baz"])
    c = nauert.PitchedQEvent(abjad.Offset(9999), [0])
    d = nauert.PitchedQEvent(abjad.Offset(1000), [0, 1, 4])
    assert a != b
    assert a != c
    assert a != d


def test_PitchedQEvent___eq___03():
    a = nauert.TerminalQEvent(abjad.Offset(100))
    b = nauert.PitchedQEvent(abjad.Offset(100), [0])
    c = nauert.SilentQEvent(abjad.Offset(100))
    assert a != b
    assert a != c


def test_SilentQEvent___init___01():
    q_event = nauert.SilentQEvent(abjad.Offset(130))
    assert q_event.offset == abjad.Offset(130)
    assert q_event.attachments == ()


def test_SilentQEvent___init___02():
    attachments = ["foo", "bar", "baz"]
    q_event = nauert.SilentQEvent(abjad.Offset(155, 7), attachments=attachments)
    assert q_event.offset == abjad.Offset(155, 7)
    assert q_event.attachments == ("foo", "bar", "baz")


def test_SilentQEvent___eq___01():
    a = nauert.SilentQEvent(abjad.Offset(1000))
    b = nauert.SilentQEvent(abjad.Offset(1000))
    assert a == b


def test_SilentQEvent___eq___02():
    a = nauert.SilentQEvent(abjad.Offset(1000))
    b = nauert.SilentQEvent(abjad.Offset(1000), ["foo", "bar", "baz"])
    c = nauert.SilentQEvent(abjad.Offset(9999))
    assert a != b
    assert a != c


def test_TerminalQEvent___init___01():
    q_event = nauert.TerminalQEvent(abjad.Offset(154))
    assert q_event.offset == abjad.Offset(154)


def test_TerminalQEvent___eq___01():
    a = nauert.TerminalQEvent(abjad.Offset(1000))
    b = nauert.TerminalQEvent(abjad.Offset(1000))
    assert a == b


def test_TerminalQEvent___eq___02():
    a = nauert.TerminalQEvent(abjad.Offset(1000))
    b = nauert.TerminalQEvent(abjad.Offset(9000))
    assert a != b


def test_TerminalQEvent___eq___03():
    a = nauert.TerminalQEvent(abjad.Offset(100))
    b = nauert.PitchedQEvent(abjad.Offset(100), [0])
    c = nauert.SilentQEvent(abjad.Offset(100))
    assert a != b
    assert a != c


def test_QEvent_from_offset_pitches_attachments():
    q_event = nauert.QEvent.from_offset_pitches_attachments(
        abjad.Offset(100), 1, ("foo",)
    )
    assert isinstance(q_event, nauert.PitchedQEvent)
    assert q_event.offset == 100
    assert q_event.pitches == (abjad.NamedPitch(1),)
    assert q_event.attachments == ("foo",)


def test_QEvent_from_offset_pitches_attachments_with_incorrectly_typed_pitches():
    with pytest.raises(TypeError):
        nauert.QEvent.from_offset_pitches_attachments(100, decimal.Decimal(0), ("foo",))



import abjad

import nauert


def test_BeatwiseQSchema___getitem___01():
    schema = nauert.BeatwiseQSchema()
    assert (
        schema[0]
        == schema[1]
        == schema[2]
        == {
            "beatspan": abjad.Duration(1, 4),
            "search_tree": nauert.UnweightedSearchTree(),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 60),
        }
    )


def test_BeatwiseQSchema___getitem___02():
    item_a = nauert.BeatwiseQSchemaItem(
        search_tree=nauert.UnweightedSearchTree({2: None})
    )
    metronome_mark = abjad.MetronomeMark(abjad.Duration(1, 4), 76)
    item_b = nauert.BeatwiseQSchemaItem(tempo=metronome_mark)
    item_c = nauert.BeatwiseQSchemaItem(
        beatspan=abjad.Duration(1, 8),
        search_tree=nauert.UnweightedSearchTree({5: None}),
    )
    schema = nauert.BeatwiseQSchema(
        {2: item_a, 4: item_b, 7: item_c},
        beatspan=abjad.Duration(1, 32),
        search_tree=nauert.UnweightedSearchTree({3: None}),
        tempo=abjad.MetronomeMark(abjad.Duration(1, 16), 36),
    )
    assert (
        schema[0]
        == schema[1]
        == {
            "beatspan": abjad.Duration(1, 32),
            "search_tree": nauert.UnweightedSearchTree({3: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 16), 36),
        }
    )
    assert (
        schema[2]
        == schema[3]
        == {
            "beatspan": abjad.Duration(1, 32),
            "search_tree": nauert.UnweightedSearchTree({2: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 16), 36),
        }
    )
    assert (
        schema[4]
        == schema[5]
        == schema[6]
        == {
            "beatspan": abjad.Duration(1, 32),
            "search_tree": nauert.UnweightedSearchTree({2: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 76),
        }
    )
    assert (
        schema[7]
        == schema[8]
        == schema[1000]
        == {
            "beatspan": abjad.Duration(1, 8),
            "search_tree": nauert.UnweightedSearchTree({5: None}),
            "tempo": abjad.MetronomeMark(abjad.Duration(1, 4), 76),
        }
    )
